"""Add CI/CD, deployment, and incident tables

Revision ID: c3a7b2d4e5f6
Revises: 9f2f70b48ab5
Create Date: 2025-09-17 12:00:00.000000

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op


revision: str = "c3a7b2d4e5f6"
down_revision: Union[str, None] = "9f2f70b48ab5"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    op.create_table(
        "ci_pipeline_runs",
        sa.Column("repo_id", sa.UUID(), nullable=False),
        sa.Column("run_id", sa.Text(), nullable=False),
        sa.Column("status", sa.Text(), nullable=True),
        sa.Column("queued_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("started_at", sa.DateTime(timezone=True), nullable=False),
        sa.Column("finished_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column(
            "_mergestat_synced_at",
            sa.DateTime(timezone=True),
            nullable=False,
            comment="timestamp when record was synced into the MergeStat database",
        ),
        sa.ForeignKeyConstraint(["repo_id"], ["repos.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("repo_id", "run_id"),
    )

    op.create_table(
        "deployments",
        sa.Column("repo_id", sa.UUID(), nullable=False),
        sa.Column("deployment_id", sa.Text(), nullable=False),
        sa.Column("status", sa.Text(), nullable=True),
        sa.Column("environment", sa.Text(), nullable=True),
        sa.Column("started_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("finished_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("deployed_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("merged_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("pull_request_number", sa.Integer(), nullable=True),
        sa.Column(
            "_mergestat_synced_at",
            sa.DateTime(timezone=True),
            nullable=False,
            comment="timestamp when record was synced into the MergeStat database",
        ),
        sa.ForeignKeyConstraint(["repo_id"], ["repos.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("repo_id", "deployment_id"),
    )

    op.create_table(
        "incidents",
        sa.Column("repo_id", sa.UUID(), nullable=False),
        sa.Column("incident_id", sa.Text(), nullable=False),
        sa.Column("status", sa.Text(), nullable=True),
        sa.Column("started_at", sa.DateTime(timezone=True), nullable=False),
        sa.Column("resolved_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column(
            "_mergestat_synced_at",
            sa.DateTime(timezone=True),
            nullable=False,
            comment="timestamp when record was synced into the MergeStat database",
        ),
        sa.ForeignKeyConstraint(["repo_id"], ["repos.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("repo_id", "incident_id"),
    )


def downgrade() -> None:
    op.drop_table("incidents")
    op.drop_table("deployments")
    op.drop_table("ci_pipeline_runs")



from __future__ import annotations

import os
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Mapping, Optional, Tuple

import yaml

DEFAULT_TEAM_MAPPING_PATH = Path("config/team_mapping.yaml")


def _norm_key(value: str) -> str:
    return " ".join((value or "").strip().lower().split())


@dataclass(frozen=True)
class TeamResolver:
    member_to_team: Mapping[str, Tuple[str, str]]  # member_identity -> (team_id, team_name)

    def resolve(self, identity: Optional[str]) -> Tuple[Optional[str], Optional[str]]:
        if not identity:
            return None, None
        key = _norm_key(identity)
        team = self.member_to_team.get(key)
        if not team:
            return None, None
        return team[0], team[1]


def load_team_resolver(path: Optional[Path] = None) -> TeamResolver:
    raw_path = os.getenv("TEAM_MAPPING_PATH")
    if raw_path:
        path = Path(raw_path)
    path = path or DEFAULT_TEAM_MAPPING_PATH

    try:
        with path.open("r", encoding="utf-8") as handle:
            payload = yaml.safe_load(handle) or {}
    except FileNotFoundError:
        payload = {}

    member_to_team: Dict[str, Tuple[str, str]] = {}
    for entry in payload.get("teams") or []:
        team_id = str(entry.get("team_id") or "").strip()
        team_name = str(entry.get("team_name") or team_id).strip()
        if not team_id:
            continue
        for member in entry.get("members") or []:
            key = _norm_key(str(member))
            if not key:
                continue
            member_to_team[key] = (team_id, team_name)

    return TeamResolver(member_to_team=member_to_team)




from __future__ import annotations

import uuid
from datetime import datetime, timezone

from processors.local import infer_merged_pull_requests_from_commits


class DummyPerson:
    def __init__(self, name: str, email: str):
        self.name = name
        self.email = email


class DummyCommit:
    def __init__(self, message: str, committed_at: datetime):
        self.hexsha = "deadbeef"
        self.message = message
        self.author = DummyPerson("Alice", "alice@example.com")
        self.committer = DummyPerson("Alice", "alice@example.com")
        self.committed_datetime = committed_at
        self.parents = [object(), object()]


def test_infer_github_merged_pr_from_merge_commit_message():
    now = datetime.now(timezone.utc)
    message = """Merge pull request #123 from org/branch

Add feature X

Some body text
"""
    prs = infer_merged_pull_requests_from_commits(
        [DummyCommit(message, now)],
        repo_id=uuid.uuid4(),
    )
    assert len(prs) == 1
    assert prs[0].number == 123
    assert prs[0].state == "merged"
    assert prs[0].title == "Add feature X"
    assert prs[0].created_at is not None
    assert prs[0].merged_at is not None


def test_infer_gitlab_merged_mr_from_trailer():
    now = datetime.now(timezone.utc)
    message = """Merge branch 'feature' into 'main'

My MR title

See merge request group/project!45
"""
    prs = infer_merged_pull_requests_from_commits(
        [DummyCommit(message, now)],
        repo_id=uuid.uuid4(),
    )
    assert len(prs) == 1
    assert prs[0].number == 45
    assert prs[0].state == "merged"
    assert prs[0].title == "My MR title"



from __future__ import annotations

from datetime import datetime, timezone
from typing import List, Optional

import uuid
from sqlalchemy import Column, Text, DateTime, JSON
from models.git import Base, GUID

class Team(Base):
    __tablename__ = "teams"

    id = Column(Text, primary_key=True, comment="Unique team identifier (slug)")
    team_uuid = Column(GUID, unique=True, default=uuid.uuid4, comment="Internal unique identifier")
    name = Column(Text, nullable=False, comment="Team display name")
    description = Column(Text, nullable=True, comment="Team description")
    members = Column(JSON, default=list, comment="List of member identities")
    updated_at = Column(
        DateTime(timezone=True), 
        default=lambda: datetime.now(timezone.utc), 
        onupdate=lambda: datetime.now(timezone.utc)
    )

    def __init__(
        self, 
        id: str, 
        name: str, 
        description: Optional[str] = None, 
        members: Optional[List[str]] = None, 
        updated_at: Optional[datetime] = None,
        team_uuid: Optional[uuid.UUID] = None
    ):
        self.id = id
        self.team_uuid = team_uuid or uuid.uuid4()
        self.name = name
        self.description = description
        self.members = members or []
        self.updated_at = updated_at or datetime.now(timezone.utc)


"""
Integration example showing how to use connectors with existing storage system.

This example demonstrates how to retrieve data from GitHub/GitLab and
store it in the existing database using the storage system.
"""

import asyncio
import os
import sys
from datetime import datetime, timezone

# Add parent directory to path for imports
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from connectors import GitHubConnector, GitLabConnector
from models.git import GitCommit, GitCommitStat, Repo
from storage import SQLAlchemyStore


async def github_to_storage_example():
    """Example: Retrieve GitHub data and store in database."""
    token = os.getenv("GITHUB_TOKEN")
    if not token:
        print("GITHUB_TOKEN not set, skipping GitHub example")
        return

    # Database connection
    db_conn = os.getenv("DB_CONN_STRING", "sqlite+aiosqlite:///./test_integration.db")

    print("=== GitHub to Storage Integration ===\n")

    # Initialize connector
    connector = GitHubConnector(token=token)

    try:
        # Get a repository
        print("Fetching repositories...")
        repos = connector.list_repositories(max_repos=1)

        if not repos:
            print("No repositories found")
            return

        repo = repos[0]
        print(f"Using repository: {repo.full_name}\n")

        # Store repository in database
        async with SQLAlchemyStore(db_conn) as store:
            # Create Repo object for database
            db_repo = Repo(
                repo_path=None,  # Not a local repo
                repo=repo.full_name,
                settings={"source": "github", "repo_id": repo.id},
                tags=["github", repo.language] if repo.language else ["github"],
            )

            print(f"Storing repository: {db_repo.repo}")
            await store.insert_repo(db_repo)
            print(f"Repository stored with ID: {db_repo.id}\n")

            # Get and store commits
            print("Fetching commits...")
            _, repo_name = repo.full_name.split("/")

            # For demo purposes, we'll use PyGithub to get commits
            gh_repo = connector.github.get_repo(repo.full_name)
            commits = list(gh_repo.get_commits()[:5])  # Get first 5 commits

            commit_objects = []
            for commit in commits:
                git_commit = GitCommit(
                    repo_id=db_repo.id,
                    hash=commit.sha,
                    message=commit.commit.message,
                    author_name=(
                        commit.commit.author.name if commit.commit.author else "Unknown"
                    ),
                    author_email=(
                        commit.commit.author.email if commit.commit.author else ""
                    ),
                    author_when=(
                        commit.commit.author.date
                        if commit.commit.author
                        else datetime.now(timezone.utc)
                    ),
                    committer_name=(
                        commit.commit.committer.name
                        if commit.commit.committer
                        else "Unknown"
                    ),
                    committer_email=(
                        commit.commit.committer.email if commit.commit.committer else ""
                    ),
                    committer_when=(
                        commit.commit.committer.date
                        if commit.commit.committer
                        else datetime.now(timezone.utc)
                    ),
                    parents=len(commit.parents),
                )
                commit_objects.append(git_commit)
                # Extract first line of commit message
                first_line = commit.commit.message.split("\n")[0][:50]
                print(f"  - {commit.sha[:8]}: {first_line}")

            await store.insert_git_commit_data(commit_objects)
            print(f"\nStored {len(commit_objects)} commits")

            # Get and store commit stats
            print("\nFetching commit stats...")
            stats_objects = []
            for commit in commits[:3]:  # Just first 3 for demo
                for file in commit.files:
                    stat = GitCommitStat(
                        repo_id=db_repo.id,
                        commit_hash=commit.sha,
                        file_path=file.filename,
                        additions=file.additions,
                        deletions=file.deletions,
                        old_file_mode="unknown",
                        new_file_mode="unknown",
                    )
                    stats_objects.append(stat)

            await store.insert_git_commit_stats(stats_objects)
            print(f"Stored {len(stats_objects)} commit stats")

    except Exception as e:
        print(f"Error: {e}")
        import traceback

        traceback.print_exc()
    finally:
        connector.close()


async def gitlab_to_storage_example():
    """Example: Retrieve GitLab data and store in database."""
    token = os.getenv("GITLAB_TOKEN")
    if not token:
        print("GITLAB_TOKEN not set, skipping GitLab example")
        return

    # Database connection
    db_conn = os.getenv("DB_CONN_STRING", "sqlite+aiosqlite:///./test_integration.db")

    print("\n\n=== GitLab to Storage Integration ===\n")

    # Initialize connector
    connector = GitLabConnector(private_token=token)

    try:
        # Get a project
        print("Fetching projects...")
        projects = connector.list_projects(max_projects=1)

        if not projects:
            print("No projects found")
            return

        project = projects[0]
        print(f"Using project: {project.full_name}\n")

        # Store project in database
        async with SQLAlchemyStore(db_conn) as store:
            # Create Repo object for database
            db_repo = Repo(
                repo_path=None,  # Not a local repo
                repo=project.full_name,
                settings={"source": "gitlab", "project_id": project.id},
                tags=["gitlab"],
            )

            print(f"Storing project: {db_repo.repo}")
            await store.insert_repo(db_repo)
            print(f"Project stored with ID: {db_repo.id}\n")

            # Get and store commits
            print("Fetching commits...")
            gl_project = connector.gitlab.projects.get(project.id)
            commits = gl_project.commits.list(per_page=5, get_all=False)

            commit_objects = []
            for commit in commits:
                git_commit = GitCommit(
                    repo_id=db_repo.id,
                    hash=commit.id,
                    message=commit.message,
                    author_name=(
                        commit.author_name
                        if hasattr(commit, "author_name")
                        else "Unknown"
                    ),
                    author_email=(
                        commit.author_email if hasattr(commit, "author_email") else ""
                    ),
                    author_when=(
                        datetime.fromisoformat(
                            commit.authored_date.replace("Z", "+00:00")
                        )
                        if hasattr(commit, "authored_date")
                        else datetime.now(timezone.utc)
                    ),
                    committer_name=(
                        commit.committer_name
                        if hasattr(commit, "committer_name")
                        else "Unknown"
                    ),
                    committer_email=(
                        commit.committer_email
                        if hasattr(commit, "committer_email")
                        else ""
                    ),
                    committer_when=(
                        datetime.fromisoformat(
                            commit.committed_date.replace("Z", "+00:00")
                        )
                        if hasattr(commit, "committed_date")
                        else datetime.now(timezone.utc)
                    ),
                    parents=(
                        len(commit.parent_ids) if hasattr(commit, "parent_ids") else 0
                    ),
                )
                commit_objects.append(git_commit)
                # Extract first line of commit message
                first_line = commit.message.split("\n")[0][:50]
                print(f"  - {commit.id[:8]}: {first_line}")

            await store.insert_git_commit_data(commit_objects)
            print(f"\nStored {len(commit_objects)} commits")

    except Exception as e:
        print(f"Error: {e}")
        import traceback

        traceback.print_exc()
    finally:
        connector.close()


async def main():
    """Main function."""
    print("Integration Examples: Connectors + Storage\n")
    print("=" * 60)

    # Run GitHub example
    await github_to_storage_example()

    # Run GitLab example
    await gitlab_to_storage_example()

    print("\n" + "=" * 60)
    print("Integration examples completed!")
    print(
        "\nNote: Set GITHUB_TOKEN and GITLAB_TOKEN environment variables to run examples."
    )


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

import os
import re
from datetime import datetime, timezone
from typing import Any, List, Optional, Tuple

from models.work_items import WorkItem, WorkItemStatusTransition
from providers.identity import IdentityResolver
from providers.status_mapping import StatusMapping


def _parse_datetime(value: Any) -> Optional[datetime]:
    if value is None:
        return None
    if isinstance(value, datetime):
        dt = value
    else:
        raw = str(value).strip()
        if not raw:
            return None
        # Jira commonly uses "+0000" offsets (no colon); normalize for fromisoformat.
        raw = raw.replace("Z", "+00:00")
        if re.search(r"[+-]\d{4}$", raw):
            raw = raw[:-2] + ":" + raw[-2:]
        try:
            dt = datetime.fromisoformat(raw)
        except Exception:
            return None

    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=timezone.utc)
    return dt.astimezone(timezone.utc)


def _get_field(issue: Any, field_name: str) -> Any:
    if isinstance(issue, dict):
        fields = issue.get("fields") if isinstance(issue.get("fields"), dict) else None
        if fields is None:
            return None
        return fields.get(field_name)
    fields = getattr(issue, "fields", None)
    if fields is None:
        return None
    return getattr(fields, field_name, None)


def _parse_sprint(value: Any) -> Tuple[Optional[str], Optional[str]]:
    """
    Jira sprint fields vary by instance. Best-effort parsing:
    - list of strings with "id=...,name=..."
    - list of dict-like objects
    - single string
    """
    if not value:
        return None, None
    sprint = None
    if isinstance(value, list):
        sprint = value[-1] if value else None
    else:
        sprint = value

    if isinstance(sprint, dict):
        sid = sprint.get("id")
        name = sprint.get("name")
        return str(sid) if sid is not None else None, str(name) if name else None

    raw = str(sprint)
    # Typical Jira string contains "id=123,name=Sprint 1,".
    sid_match = re.search(r"\bid=(\d+)\b", raw)
    name_match = re.search(r"\bname=([^,\\]]+)", raw)
    sid = sid_match.group(1) if sid_match else None
    name = name_match.group(1).strip() if name_match else None
    return sid, name


def jira_issue_to_work_item(
    *,
    issue: Any,
    status_mapping: StatusMapping,
    identity: IdentityResolver,
    repo_id: Optional[Any] = None,
    story_points_field: Optional[str] = None,
    sprint_field: Optional[str] = None,
    epic_link_field: Optional[str] = None,
) -> Tuple[WorkItem, List[WorkItemStatusTransition]]:
    """
    Normalize a Jira issue into a WorkItem and status transitions.

    `story_points_field`, `sprint_field`, and `epic_link_field` are instance-specific.
    If not provided, environment variables are used:
    - JIRA_STORY_POINTS_FIELD
    - JIRA_SPRINT_FIELD
    - JIRA_EPIC_LINK_FIELD
    """
    story_points_field = story_points_field or os.getenv("JIRA_STORY_POINTS_FIELD")
    sprint_field = sprint_field or os.getenv("JIRA_SPRINT_FIELD") or "customfield_10020"
    epic_link_field = epic_link_field or os.getenv("JIRA_EPIC_LINK_FIELD")

    key = (issue.get("key") if isinstance(issue, dict) else getattr(issue, "key", None)) or ""
    work_item_id = f"jira:{key}"

    project = _get_field(issue, "project")
    if isinstance(project, dict):
        project_key = project.get("key")
        project_id = project.get("id")
    else:
        project_key = getattr(project, "key", None) if project else None
        project_id = getattr(project, "id", None) if project else None

    title = _get_field(issue, "summary") or ""

    status_obj = _get_field(issue, "status")
    if isinstance(status_obj, dict):
        status_raw = status_obj.get("name")
        status_category_key = ((status_obj.get("statusCategory") or {}) if isinstance(status_obj.get("statusCategory"), dict) else {}).get("key")
    else:
        status_raw = getattr(status_obj, "name", None) if status_obj else None
        status_category_key = getattr(getattr(status_obj, "statusCategory", None), "key", None) if status_obj else None

    issue_type_obj = _get_field(issue, "issuetype")
    if isinstance(issue_type_obj, dict):
        type_raw = issue_type_obj.get("name")
    else:
        type_raw = getattr(issue_type_obj, "name", None) if issue_type_obj else None

    labels = list(_get_field(issue, "labels") or [])

    normalized_status = status_mapping.normalize_status(
        provider="jira",
        status_raw=status_raw,
        labels=labels,
    )
    # Jira statusCategory=done is a strong hint that the issue is completed even if the status name is custom.
    if normalized_status in {"unknown", "todo", "in_progress", "in_review", "blocked", "backlog"}:
        if str(status_category_key or "").lower() == "done":
            normalized_status = "done"
    normalized_type = status_mapping.normalize_type(
        provider="jira",
        type_raw=type_raw,
        labels=labels,
    )

    assignees: List[str] = []
    assignee_obj = _get_field(issue, "assignee")
    if assignee_obj is not None:
        assignees.append(
            identity.resolve(
                provider="jira",
                email=getattr(assignee_obj, "emailAddress", None),
                account_id=getattr(assignee_obj, "accountId", None),
                display_name=getattr(assignee_obj, "displayName", None),
            )
        )

    reporter_obj = _get_field(issue, "reporter")
    reporter = None
    if reporter_obj is not None:
        reporter = identity.resolve(
            provider="jira",
            email=getattr(reporter_obj, "emailAddress", None),
            account_id=getattr(reporter_obj, "accountId", None),
            display_name=getattr(reporter_obj, "displayName", None),
        )

    created_at = _parse_datetime(_get_field(issue, "created")) or datetime.now(timezone.utc)
    updated_at = _parse_datetime(_get_field(issue, "updated")) or created_at
    closed_at = _parse_datetime(_get_field(issue, "resolutiondate"))

    url = None
    if isinstance(issue, dict):
        url = issue.get("self")
    elif hasattr(issue, "self"):
        url = getattr(issue, "self", None)

    story_points = None
    if story_points_field:
        raw_points = _get_field(issue, story_points_field)
        try:
            story_points = float(raw_points) if raw_points is not None else None
        except Exception:
            story_points = None

    sprint_id = None
    sprint_name = None
    if sprint_field:
        sprint_id, sprint_name = _parse_sprint(_get_field(issue, sprint_field))

    parent_id = None
    parent_obj = _get_field(issue, "parent")
    if parent_obj is not None:
        parent_id = f"jira:{getattr(parent_obj, 'key', None) or ''}" or None

    epic_id = None
    if epic_link_field:
        epic_val = _get_field(issue, epic_link_field)
        if epic_val:
            epic_id = f"jira:{str(epic_val)}"

    # Changelog transitions for started/completed derivation.
    transitions: List[WorkItemStatusTransition] = []
    if isinstance(issue, dict):
        changelog = issue.get("changelog") if isinstance(issue.get("changelog"), dict) else None
        histories = changelog.get("histories") if changelog else None
    else:
        changelog = getattr(issue, "changelog", None)
        histories = getattr(changelog, "histories", None) if changelog else None
    if histories:
        # Jira returns newest-first sometimes; sort by created timestamp.
        def _hist_dt(h: Any) -> datetime:
            created = h.get("created") if isinstance(h, dict) else getattr(h, "created", None)
            return _parse_datetime(created) or datetime.min.replace(
                tzinfo=timezone.utc
            )

        for hist in sorted(list(histories), key=_hist_dt):
            occurred_at = _parse_datetime(hist.get("created") if isinstance(hist, dict) else getattr(hist, "created", None)) or created_at
            author_obj = hist.get("author") if isinstance(hist, dict) else getattr(hist, "author", None)
            actor = None
            if author_obj is not None:
                actor = identity.resolve(
                    provider="jira",
                    email=author_obj.get("emailAddress") if isinstance(author_obj, dict) else getattr(author_obj, "emailAddress", None),
                    account_id=author_obj.get("accountId") if isinstance(author_obj, dict) else getattr(author_obj, "accountId", None),
                    display_name=author_obj.get("displayName") if isinstance(author_obj, dict) else getattr(author_obj, "displayName", None),
                )
            items = hist.get("items") if isinstance(hist, dict) else getattr(hist, "items", None)
            for item in items or []:
                field_name = item.get("field") if isinstance(item, dict) else getattr(item, "field", "")
                if str(field_name or "").lower() != "status":
                    continue
                from_raw = item.get("fromString") if isinstance(item, dict) else getattr(item, "fromString", None)
                to_raw = item.get("toString") if isinstance(item, dict) else getattr(item, "toString", None)
                from_norm = status_mapping.normalize_status(
                    provider="jira", status_raw=from_raw, labels=labels
                )
                to_norm = status_mapping.normalize_status(
                    provider="jira", status_raw=to_raw, labels=labels
                )
                transitions.append(
                    WorkItemStatusTransition(
                        work_item_id=work_item_id,
                        provider="jira",
                        occurred_at=occurred_at,
                        from_status_raw=from_raw,
                        to_status_raw=to_raw,
                        from_status=from_norm,
                        to_status=to_norm,
                        actor=actor,
                    )
                )

    started_at = None
    completed_at = None
    for t in transitions:
        if started_at is None and t.to_status == "in_progress":
            started_at = t.occurred_at
        if completed_at is None and t.to_status in {"done", "canceled"}:
            completed_at = t.occurred_at
            break

    # Fallback for closed issues with no changelog.
    if completed_at is None and normalized_status in {"done", "canceled"}:
        completed_at = closed_at or updated_at

    work_item = WorkItem(
        work_item_id=work_item_id,
        provider="jira",
        repo_id=repo_id,
        project_key=str(project_key) if project_key else None,
        project_id=str(project_id) if project_id else None,
        title=str(title),
        type=normalized_type,
        status=normalized_status,
        status_raw=str(status_raw) if status_raw else None,
        assignees=[a for a in assignees if a and a != "unknown"],
        reporter=reporter if reporter and reporter != "unknown" else None,
        created_at=created_at,
        updated_at=updated_at,
        started_at=started_at,
        completed_at=completed_at,
        closed_at=closed_at,
        labels=[str(l) for l in labels if l],
        story_points=story_points,
        sprint_id=sprint_id,
        sprint_name=sprint_name,
        parent_id=parent_id,
        epic_id=epic_id,
        url=url,
    )
    return work_item, transitions



"""
Tests for connector exceptions.
"""

import pytest

from connectors.exceptions import (APIException, AuthenticationException,
                                   ConnectorException, NotFoundException,
                                   PaginationException, RateLimitException)


class TestExceptions:
    """Test exception classes."""

    def test_connector_exception(self):
        """Test base ConnectorException."""
        with pytest.raises(ConnectorException) as exc_info:
            raise ConnectorException("Test error")
        # Assertions outside with block to inspect captured exception
        assert "Test error" in str(exc_info.value)

    def test_rate_limit_exception(self):
        """Test RateLimitException."""
        with pytest.raises(RateLimitException) as exc_info:
            raise RateLimitException("Rate limit exceeded")
        # Assertions outside with block to inspect captured exception
        assert "Rate limit exceeded" in str(exc_info.value)
        assert isinstance(exc_info.value, ConnectorException)

    def test_authentication_exception(self):
        """Test AuthenticationException."""
        with pytest.raises(AuthenticationException) as exc_info:
            raise AuthenticationException("Auth failed")
        # Assertions outside with block to inspect captured exception
        assert "Auth failed" in str(exc_info.value)
        assert isinstance(exc_info.value, ConnectorException)

    def test_not_found_exception(self):
        """Test NotFoundException."""
        with pytest.raises(NotFoundException) as exc_info:
            raise NotFoundException("Resource not found")
        # Assertions outside with block to inspect captured exception
        assert "Resource not found" in str(exc_info.value)
        assert isinstance(exc_info.value, ConnectorException)

    def test_pagination_exception(self):
        """Test PaginationException."""
        with pytest.raises(PaginationException) as exc_info:
            raise PaginationException("Pagination error")
        # Assertions outside with block to inspect captured exception
        assert "Pagination error" in str(exc_info.value)
        assert isinstance(exc_info.value, ConnectorException)

    def test_api_exception(self):
        """Test APIException."""
        with pytest.raises(APIException) as exc_info:
            raise APIException("API error")
        # Assertions outside with block to inspect captured exception
        assert "API error" in str(exc_info.value)
        assert isinstance(exc_info.value, ConnectorException)

    def test_exception_hierarchy(self):
        """Test exception hierarchy."""
        assert issubclass(RateLimitException, ConnectorException)
        assert issubclass(AuthenticationException, ConnectorException)
        assert issubclass(NotFoundException, ConnectorException)
        assert issubclass(PaginationException, ConnectorException)
        assert issubclass(APIException, ConnectorException)

    def test_catching_base_exception(self):
        """Test catching derived exceptions with base class."""
        try:
            raise RateLimitException("Rate limit")
        except ConnectorException as e:
            assert isinstance(e, RateLimitException)

        try:
            raise AuthenticationException("Auth error")
        except ConnectorException as e:
            assert isinstance(e, AuthenticationException)



import uuid
from datetime import date, datetime, timezone
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
import pytest_asyncio
from pymongo import UpdateOne
from sqlalchemy import select, text

from models import GitBlame, GitCommit, GitCommitStat, GitFile, Repo
from models.git import Base
from storage import (
    ClickHouseStore,
    MongoStore,
    SQLAlchemyStore,
    create_store,
    detect_db_type,
    model_to_dict,
)


class TestDetectDbType:
    """Tests for the detect_db_type function."""

    def test_detect_clickhouse(self):
        """Test detection of ClickHouse connection strings."""
        assert detect_db_type("clickhouse://localhost:8123/default") == "clickhouse"
        assert (
            detect_db_type("clickhouse+http://localhost:8123/default") == "clickhouse"
        )
        assert (
            detect_db_type("clickhouse+https://localhost:8443/default") == "clickhouse"
        )

    def test_detect_mongodb(self):
        """Test detection of MongoDB connection strings."""
        assert detect_db_type("mongodb://localhost:27017/mydb") == "mongo"
        assert (
            detect_db_type("mongodb+srv://user:pass@cluster.mongodb.net/db") == "mongo"
        )

    def test_detect_postgresql(self):
        """Test detection of PostgreSQL connection strings."""
        assert detect_db_type("postgresql://localhost/mydb") == "postgres"
        assert detect_db_type("postgres://localhost/mydb") == "postgres"
        assert detect_db_type("postgresql+asyncpg://localhost/mydb") == "postgres"

    def test_detect_sqlite(self):
        """Test detection of SQLite connection strings."""
        assert detect_db_type("sqlite:///test.db") == "sqlite"
        assert detect_db_type("sqlite+aiosqlite:///:memory:") == "sqlite"

    def test_detect_case_insensitive(self):
        """Test that detection is case-insensitive."""
        assert detect_db_type("PostgreSQL://localhost/mydb") == "postgres"
        assert detect_db_type("POSTGRESQL://localhost/mydb") == "postgres"
        assert detect_db_type("MongoDB://localhost:27017/mydb") == "mongo"
        assert detect_db_type("MONGODB://localhost:27017/mydb") == "mongo"
        assert detect_db_type("SQLite:///test.db") == "sqlite"
        assert detect_db_type("SQLITE:///test.db") == "sqlite"

    def test_detect_empty_string_raises(self):
        """Test that empty connection string raises ValueError."""
        with pytest.raises(ValueError, match="Connection string is required"):
            detect_db_type("")

    def test_detect_unknown_raises(self):
        """Test that unknown connection string raises ValueError."""
        with pytest.raises(ValueError, match="Could not detect database type"):
            detect_db_type("unknown://localhost/mydb")


class TestCreateStore:
    """Tests for the create_store factory function."""

    def test_create_mongo_store(self):
        """Test creation of MongoStore from connection string."""
        store = create_store("mongodb://localhost:27017/mydb")
        assert isinstance(store, MongoStore)

    def test_create_sqlalchemy_store_postgres(self):
        """Test creation of SQLAlchemyStore for PostgreSQL."""
        store = create_store("postgresql+asyncpg://localhost/mydb")
        assert isinstance(store, SQLAlchemyStore)

    def test_create_sqlalchemy_store_sqlite(self):
        """Test creation of SQLAlchemyStore for SQLite."""
        store = create_store("sqlite+aiosqlite:///:memory:")
        assert isinstance(store, SQLAlchemyStore)

    def test_create_clickhouse_store(self):
        """Test creation of ClickHouseStore from connection string."""
        store = create_store("clickhouse://localhost:8123/default")
        assert isinstance(store, ClickHouseStore)

    def test_create_store_with_explicit_db_type(self):
        """Test creation with explicit db_type overriding auto-detection."""
        # Force PostgreSQL even though URL looks like SQLite
        store = create_store("postgresql+asyncpg://localhost/mydb", db_type="postgres")
        assert isinstance(store, SQLAlchemyStore)

    def test_create_store_unsupported_type_raises(self):
        """Test that unsupported db_type raises ValueError."""
        with pytest.raises(ValueError, match="Unsupported database type"):
            create_store("postgresql://localhost/mydb", db_type="oracle")


def test_model_to_dict_serializes_uuid_and_fields(repo_uuid):
    blame = GitBlame(
        repo_id=repo_uuid,
        path="file.txt",
        line_no=1,
        author_email="author@example.com",
        author_name="Author",
        author_when=datetime(2024, 1, 1, tzinfo=timezone.utc),
        commit_hash="abc123",
        line="content",
    )

    doc = model_to_dict(blame)

    assert doc["repo_id"] == str(repo_uuid)
    assert doc["path"] == "file.txt"
    assert doc["line_no"] == 1
    assert doc["commit_hash"] == "abc123"


@pytest.fixture
def test_db_url():
    """Return a SQLite in-memory database URL for testing."""
    return "sqlite+aiosqlite:///:memory:"


@pytest_asyncio.fixture
async def sqlalchemy_store(test_db_url):
    """Create a SQLAlchemyStore instance with an in-memory database."""
    store = SQLAlchemyStore(test_db_url)

    # Create tables
    async with store.engine.begin() as conn:
        await conn.run_sync(Base.metadata.create_all)

    yield store

    # Cleanup
    async with store.engine.begin() as conn:
        await conn.run_sync(Base.metadata.drop_all)
    await store.engine.dispose()


@pytest.mark.asyncio
async def test_sqlalchemy_store_context_manager(test_db_url):
    """Test that SQLAlchemyStore can be used as an async context manager."""
    store = SQLAlchemyStore(test_db_url)

    # Create tables first
    async with store.engine.begin() as conn:
        await conn.run_sync(Base.metadata.create_all)

    async with store as s:
        assert s.session is not None
        assert s == store

    # After exiting context, session should be closed
    # Note: session is closed but not None
    await store.engine.dispose()


@pytest.mark.asyncio
async def test_sqlalchemy_store_insert_repo(sqlalchemy_store):
    """Test inserting a repository into the database."""
    test_repo = Repo(
        id=uuid.uuid4(),
        repo="https://github.com/test/repo.git",
        ref="main",
        settings={},
        tags=[],
    )

    async with sqlalchemy_store as store:
        await store.insert_repo(test_repo)

        # Verify the repo was inserted
        result = await store.session.execute(
            select(Repo).where(Repo.id == test_repo.id)
        )
        saved_repo = result.scalar_one_or_none()

        assert saved_repo is not None
        assert saved_repo.id == test_repo.id
        assert saved_repo.repo == "https://github.com/test/repo.git"
        assert saved_repo.ref == "main"


@pytest.mark.asyncio
async def test_sqlalchemy_store_insert_repo_duplicate(sqlalchemy_store):
    """Test that inserting a duplicate repo does not cause an error."""
    test_repo = Repo(
        id=uuid.uuid4(),
        repo="https://github.com/test/repo.git",
        ref="main",
        settings={},
        tags=[],
    )

    async with sqlalchemy_store as store:
        # Insert the repo twice
        await store.insert_repo(test_repo)
        await store.insert_repo(test_repo)

        # Verify only one repo exists
        result = await store.session.execute(
            select(Repo).where(Repo.id == test_repo.id)
        )
        repos = result.scalars().all()

        assert len(repos) == 1


@pytest.mark.asyncio
async def test_sqlalchemy_store_get_complexity_snapshots_latest_for_repo(sqlalchemy_store):
    repo_id = uuid.uuid4()
    other_repo_id = uuid.uuid4()
    computed_at = datetime(2025, 1, 10, tzinfo=timezone.utc).isoformat()

    async with sqlalchemy_store as store:
        await store.insert_repo(Repo(id=repo_id, repo="owner/repo"))
        await store.insert_repo(Repo(id=other_repo_id, repo="owner/other"))

        await store.session.execute(
            text(
                """
                CREATE TABLE IF NOT EXISTS file_complexity_snapshots (
                  repo_id TEXT NOT NULL,
                  as_of_day TEXT NOT NULL,
                  ref TEXT NOT NULL,
                  file_path TEXT NOT NULL,
                  language TEXT,
                  loc INTEGER NOT NULL,
                  functions_count INTEGER NOT NULL,
                  cyclomatic_total INTEGER NOT NULL,
                  cyclomatic_avg REAL NOT NULL,
                  high_complexity_functions INTEGER NOT NULL,
                  very_high_complexity_functions INTEGER NOT NULL,
                  computed_at TEXT NOT NULL,
                  PRIMARY KEY (repo_id, as_of_day, file_path)
                )
                """
            )
        )
        await store.session.commit()

        # repo_id: snapshots on two different days
        await store.session.execute(
            text(
                """
                INSERT INTO file_complexity_snapshots
                (repo_id, as_of_day, ref, file_path, language, loc, functions_count,
                 cyclomatic_total, cyclomatic_avg, high_complexity_functions,
                 very_high_complexity_functions, computed_at)
                VALUES
                (:repo_id, '2025-01-01', 'main', 'a.py', 'python', 10, 1, 5, 5.0, 0, 0, :computed_at),
                (:repo_id, '2025-01-03', 'main', 'a.py', 'python', 12, 1, 8, 8.0, 0, 0, :computed_at),
                (:repo_id, '2025-01-03', 'main', 'b.py', 'python', 20, 2, 3, 1.5, 0, 0, :computed_at)
                """
            ),
            {"repo_id": str(repo_id), "computed_at": computed_at},
        )
        # other_repo_id: only one day
        await store.session.execute(
            text(
                """
                INSERT INTO file_complexity_snapshots
                (repo_id, as_of_day, ref, file_path, language, loc, functions_count,
                 cyclomatic_total, cyclomatic_avg, high_complexity_functions,
                 very_high_complexity_functions, computed_at)
                VALUES
                (:repo_id, '2025-01-02', 'main', 'x.py', 'python', 1, 1, 1, 1.0, 0, 0, :computed_at)
                """
            ),
            {"repo_id": str(other_repo_id), "computed_at": computed_at},
        )
        await store.session.commit()

        snaps = await store.get_complexity_snapshots(
            as_of_day=date(2025, 1, 5),
            repo_id=repo_id,
        )
        assert {s.file_path for s in snaps} == {"a.py", "b.py"}
        assert {s.as_of_day for s in snaps} == {date(2025, 1, 3)}


@pytest.mark.asyncio
async def test_sqlalchemy_store_get_complexity_snapshots_latest_for_all_repos(sqlalchemy_store):
    repo_id = uuid.uuid4()
    other_repo_id = uuid.uuid4()
    computed_at = datetime(2025, 1, 10, tzinfo=timezone.utc).isoformat()

    async with sqlalchemy_store as store:
        await store.insert_repo(Repo(id=repo_id, repo="owner/repo"))
        await store.insert_repo(Repo(id=other_repo_id, repo="owner/other"))

        await store.session.execute(
            text(
                """
                CREATE TABLE IF NOT EXISTS file_complexity_snapshots (
                  repo_id TEXT NOT NULL,
                  as_of_day TEXT NOT NULL,
                  ref TEXT NOT NULL,
                  file_path TEXT NOT NULL,
                  language TEXT,
                  loc INTEGER NOT NULL,
                  functions_count INTEGER NOT NULL,
                  cyclomatic_total INTEGER NOT NULL,
                  cyclomatic_avg REAL NOT NULL,
                  high_complexity_functions INTEGER NOT NULL,
                  very_high_complexity_functions INTEGER NOT NULL,
                  computed_at TEXT NOT NULL,
                  PRIMARY KEY (repo_id, as_of_day, file_path)
                )
                """
            )
        )
        await store.session.commit()

        await store.session.execute(
            text(
                """
                INSERT INTO file_complexity_snapshots
                (repo_id, as_of_day, ref, file_path, language, loc, functions_count,
                 cyclomatic_total, cyclomatic_avg, high_complexity_functions,
                 very_high_complexity_functions, computed_at)
                VALUES
                (:repo1, '2025-01-01', 'main', 'a.py', 'python', 10, 1, 5, 5.0, 0, 0, :computed_at),
                (:repo1, '2025-01-03', 'main', 'a.py', 'python', 12, 1, 8, 8.0, 0, 0, :computed_at),
                (:repo2, '2025-01-02', 'main', 'x.py', 'python', 1, 1, 1, 1.0, 0, 0, :computed_at),
                (:repo2, '2025-01-04', 'main', 'x.py', 'python', 2, 1, 2, 2.0, 0, 0, :computed_at)
                """
            ),
            {"repo1": str(repo_id), "repo2": str(other_repo_id), "computed_at": computed_at},
        )
        await store.session.commit()

        snaps = await store.get_complexity_snapshots(as_of_day=date(2025, 1, 5))
        by_repo = {}
        for s in snaps:
            by_repo.setdefault(s.repo_id, []).append(s)

        assert {s.as_of_day for s in by_repo[repo_id]} == {date(2025, 1, 3)}
        assert {s.as_of_day for s in by_repo[other_repo_id]} == {date(2025, 1, 4)}


@pytest.mark.asyncio
async def test_sqlalchemy_store_get_work_item_user_metrics_daily(sqlalchemy_store):
    computed_at = datetime(2025, 1, 10, tzinfo=timezone.utc).isoformat()

    async with sqlalchemy_store as store:
        await store.session.execute(
            text(
                """
                CREATE TABLE IF NOT EXISTS work_item_user_metrics_daily (
                  day TEXT NOT NULL,
                  provider TEXT NOT NULL,
                  work_scope_id TEXT NOT NULL,
                  user_identity TEXT NOT NULL,
                  team_id TEXT NOT NULL,
                  team_name TEXT NOT NULL,
                  items_started INTEGER NOT NULL,
                  items_completed INTEGER NOT NULL,
                  wip_count_end_of_day INTEGER NOT NULL,
                  cycle_time_p50_hours REAL,
                  cycle_time_p90_hours REAL,
                  computed_at TEXT NOT NULL,
                  PRIMARY KEY (provider, work_scope_id, user_identity, day)
                )
                """
            )
        )
        await store.session.execute(
            text(
                """
                INSERT INTO work_item_user_metrics_daily
                (day, provider, work_scope_id, user_identity, team_id, team_name,
                 items_started, items_completed, wip_count_end_of_day,
                 cycle_time_p50_hours, cycle_time_p90_hours, computed_at)
                VALUES
                ('2025-01-05', 'github', 'gh:owner/repo', 'dev@example.com', 'team-a', 'Team A',
                 2, 1, 3, 12.0, 48.0, :computed_at)
                """
            ),
            {"computed_at": computed_at},
        )
        await store.session.commit()

        rows = await store.get_work_item_user_metrics_daily(day=date(2025, 1, 5))
        assert len(rows) == 1
        assert rows[0].user_identity == "dev@example.com"
        assert rows[0].items_completed == 1


@pytest.mark.asyncio
async def test_sqlalchemy_store_insert_git_file_data(sqlalchemy_store):
    """Test inserting git file data into the database."""
    test_repo_id = uuid.uuid4()
    test_repo = Repo(
        id=test_repo_id,
        repo="https://github.com/test/repo.git",
        ref="main",
        settings={},
        tags=[],
    )

    file_data = [
        GitFile(
            repo_id=test_repo_id,
            path="file1.txt",
            executable=False,
            contents="content1",
        ),
        GitFile(
            repo_id=test_repo_id,
            path="file2.txt",
            executable=True,
            contents="content2",
        ),
    ]

    async with sqlalchemy_store as store:
        await store.insert_repo(test_repo)
        await store.insert_git_file_data(file_data)

        # Verify files were inserted
        result = await store.session.execute(
            select(GitFile).where(GitFile.repo_id == test_repo_id)
        )
        saved_files = result.scalars().all()

        assert len(saved_files) == 2
        assert saved_files[0].path in ["file1.txt", "file2.txt"]
        assert saved_files[1].path in ["file1.txt", "file2.txt"]


@pytest.mark.asyncio
async def test_sqlalchemy_store_insert_git_file_data_upsert(sqlalchemy_store):
    """Test that inserting duplicate git file data performs an upsert."""
    test_repo_id = uuid.uuid4()
    test_repo = Repo(
        id=test_repo_id,
        repo="https://github.com/test/repo.git",
        ref="main",
        settings={},
        tags=[],
    )

    initial = [
        GitFile(
            repo_id=test_repo_id,
            path="file.txt",
            executable=False,
            contents="content1",
        )
    ]
    updated = [
        GitFile(
            repo_id=test_repo_id,
            path="file.txt",
            executable=True,
            contents="content2",
        )
    ]

    async with sqlalchemy_store as store:
        await store.insert_repo(test_repo)
        await store.insert_git_file_data(initial)
        await store.insert_git_file_data(updated)

        result = await store.session.execute(
            select(GitFile).where(GitFile.repo_id == test_repo_id)
        )
        saved_files = result.scalars().all()

        assert len(saved_files) == 1
        assert saved_files[0].path == "file.txt"
        assert saved_files[0].executable
        assert saved_files[0].contents == "content2"


@pytest.mark.asyncio
async def test_sqlalchemy_store_insert_git_file_data_empty_list(sqlalchemy_store):
    """Test that inserting an empty list does not cause an error."""
    async with sqlalchemy_store as store:
        await store.insert_git_file_data([])
        # Should not raise any error


@pytest.mark.asyncio
async def test_sqlalchemy_store_insert_git_commit_data(sqlalchemy_store):
    """Test inserting git commit data into the database."""
    test_repo_id = uuid.uuid4()
    test_repo = Repo(
        id=test_repo_id,
        repo="https://github.com/test/repo.git",
        ref="main",
        settings={},
        tags=[],
    )

    commit_data = [
        GitCommit(
            repo_id=test_repo_id,
            hash="abc123",
            message="Initial commit",
            author_name="Test Author",
            author_email="test@example.com",
            author_when=datetime(2024, 1, 1, tzinfo=timezone.utc),
            committer_name="Test Committer",
            committer_email="committer@example.com",
            committer_when=datetime(2024, 1, 1, tzinfo=timezone.utc),
            parents=0,
        ),
        GitCommit(
            repo_id=test_repo_id,
            hash="def456",
            message="Second commit",
            author_name="Test Author",
            author_email="test@example.com",
            author_when=datetime(2024, 1, 2, tzinfo=timezone.utc),
            committer_name="Test Committer",
            committer_email="committer@example.com",
            committer_when=datetime(2024, 1, 2, tzinfo=timezone.utc),
            parents=1,
        ),
    ]

    async with sqlalchemy_store as store:
        await store.insert_repo(test_repo)
        await store.insert_git_commit_data(commit_data)

        # Verify commits were inserted
        result = await store.session.execute(
            select(GitCommit).where(GitCommit.repo_id == test_repo_id)
        )
        saved_commits = result.scalars().all()

        assert len(saved_commits) == 2
        assert saved_commits[0].hash in ["abc123", "def456"]
        assert saved_commits[1].hash in ["abc123", "def456"]


@pytest.mark.asyncio
async def test_sqlalchemy_store_insert_git_commit_data_empty_list(sqlalchemy_store):
    """Test that inserting an empty list does not cause an error."""
    async with sqlalchemy_store as store:
        await store.insert_git_commit_data([])
        # Should not raise any error


@pytest.mark.asyncio
async def test_sqlalchemy_store_insert_git_commit_stats(sqlalchemy_store):
    """Test inserting git commit stats into the database."""
    test_repo_id = uuid.uuid4()
    test_repo = Repo(
        id=test_repo_id,
        repo="https://github.com/test/repo.git",
        ref="main",
        settings={},
        tags=[],
    )

    commit_stats = [
        GitCommitStat(
            repo_id=test_repo_id,
            commit_hash="abc123",
            file_path="file1.txt",
            additions=10,
            deletions=5,
            old_file_mode="100644",
            new_file_mode="100644",
        ),
        GitCommitStat(
            repo_id=test_repo_id,
            commit_hash="abc123",
            file_path="file2.txt",
            additions=20,
            deletions=3,
            old_file_mode="100644",
            new_file_mode="100755",
        ),
    ]

    async with sqlalchemy_store as store:
        await store.insert_repo(test_repo)
        await store.insert_git_commit_stats(commit_stats)

        # Verify commit stats were inserted
        result = await store.session.execute(
            select(GitCommitStat).where(GitCommitStat.repo_id == test_repo_id)
        )
        saved_stats = result.scalars().all()

        assert len(saved_stats) == 2
        assert saved_stats[0].file_path in ["file1.txt", "file2.txt"]
        assert saved_stats[1].file_path in ["file1.txt", "file2.txt"]


@pytest.mark.asyncio
async def test_sqlalchemy_store_insert_git_commit_stats_empty_list(sqlalchemy_store):
    """Test that inserting an empty list does not cause an error."""
    async with sqlalchemy_store as store:
        await store.insert_git_commit_stats([])
        # Should not raise any error


@pytest.mark.asyncio
async def test_sqlalchemy_store_insert_blame_data(sqlalchemy_store):
    """Test inserting git blame data into the database."""
    test_repo_id = uuid.uuid4()
    test_repo = Repo(
        id=test_repo_id,
        repo="https://github.com/test/repo.git",
        ref="main",
        settings={},
        tags=[],
    )

    blame_data = [
        GitBlame(
            repo_id=test_repo_id,
            path="file.txt",
            line_no=1,
            author_email="author@example.com",
            author_name="Test Author",
            author_when=datetime(2024, 1, 1, tzinfo=timezone.utc),
            commit_hash="abc123",
            line="line 1 content",
        ),
        GitBlame(
            repo_id=test_repo_id,
            path="file.txt",
            line_no=2,
            author_email="author@example.com",
            author_name="Test Author",
            author_when=datetime(2024, 1, 1, tzinfo=timezone.utc),
            commit_hash="abc123",
            line="line 2 content",
        ),
    ]

    async with sqlalchemy_store as store:
        await store.insert_repo(test_repo)
        await store.insert_blame_data(blame_data)

        # Verify blame data was inserted
        result = await store.session.execute(
            select(GitBlame).where(GitBlame.repo_id == test_repo_id)
        )
        saved_blames = result.scalars().all()

        assert len(saved_blames) == 2
        assert saved_blames[0].line_no in [1, 2]
        assert saved_blames[1].line_no in [1, 2]


@pytest.mark.asyncio
async def test_sqlalchemy_store_insert_blame_data_empty_list(sqlalchemy_store):
    """Test that inserting an empty list does not cause an error."""
    async with sqlalchemy_store as store:
        await store.insert_blame_data([])
        # Should not raise any error


@pytest.mark.asyncio
async def test_sqlalchemy_store_session_management(test_db_url):
    """Test session lifecycle management in SQLAlchemyStore."""
    store = SQLAlchemyStore(test_db_url)

    # Create tables
    async with store.engine.begin() as conn:
        await conn.run_sync(Base.metadata.create_all)

    # Before entering context, session should be None
    assert store.session is None

    async with store as s:
        # Inside context, session should be available
        assert s.session is not None

    # After exiting context, the session reference still exists but is closed
    # We can't easily test if it's closed, but we can verify engine is disposed

    await store.engine.dispose()


@pytest.mark.asyncio
async def test_sqlalchemy_store_transaction_commit(sqlalchemy_store):
    """Test that transactions are committed properly."""
    test_repo = Repo(
        id=uuid.uuid4(),
        repo="https://github.com/test/repo.git",
        ref="main",
        settings={},
        tags=[],
    )

    async with sqlalchemy_store as store:
        await store.insert_repo(test_repo)

        # Create a new session to verify the commit happened
        async with store.session_factory() as new_session:
            result = await new_session.execute(
                select(Repo).where(Repo.id == test_repo.id)
            )
            saved_repo = result.scalar_one_or_none()

            assert saved_repo is not None
            assert saved_repo.id == test_repo.id


@pytest.mark.asyncio
async def test_sqlalchemy_store_multiple_operations(sqlalchemy_store):
    """Test performing multiple operations in a single session."""
    test_repo_id = uuid.uuid4()
    test_repo = Repo(
        id=test_repo_id,
        repo="https://github.com/test/repo.git",
        ref="main",
        settings={},
        tags=[],
    )

    file_data = [
        GitFile(
            repo_id=test_repo_id,
            path="file.txt",
            executable=False,
            contents="content",
        )
    ]

    commit_data = [
        GitCommit(
            repo_id=test_repo_id,
            hash="abc123",
            message="Initial commit",
            author_name="Test Author",
            author_email="test@example.com",
            author_when=datetime(2024, 1, 1, tzinfo=timezone.utc),
            committer_name="Test Committer",
            committer_email="committer@example.com",
            committer_when=datetime(2024, 1, 1, tzinfo=timezone.utc),
            parents=0,
        )
    ]

    async with sqlalchemy_store as store:
        await store.insert_repo(test_repo)
        await store.insert_git_file_data(file_data)
        await store.insert_git_commit_data(commit_data)

        # Verify all data was inserted
        repo_result = await store.session.execute(
            select(Repo).where(Repo.id == test_repo_id)
        )
        assert repo_result.scalar_one_or_none() is not None

        file_result = await store.session.execute(
            select(GitFile).where(GitFile.repo_id == test_repo_id)
        )
        assert len(file_result.scalars().all()) == 1

        commit_result = await store.session.execute(
            select(GitCommit).where(GitCommit.repo_id == test_repo_id)
        )
        assert len(commit_result.scalars().all()) == 1


# ClickHouse Tests


@pytest.mark.asyncio
async def test_clickhouse_store_context_manager_initializes_and_creates_tables():
    mock_client = MagicMock()
    mock_client.command = MagicMock()
    mock_client.insert = MagicMock()
    mock_client.query = MagicMock(return_value=MagicMock(result_rows=[]))
    mock_client.close = MagicMock()

    # Mock filesystem operations
    mock_sql_file = MagicMock()
    mock_sql_file.read_text.return_value = "CREATE TABLE test1; CREATE TABLE test2;"

    import sys
    from types import SimpleNamespace

    get_client = MagicMock(return_value=mock_client)
    fake_clickhouse_connect = SimpleNamespace(get_client=get_client)

    with patch.dict(sys.modules, {"clickhouse_connect": fake_clickhouse_connect}), patch(
        "storage.Path"
    ) as MockPath:

        # Setup the chain: Path(__file__).resolve().parent / "migrations" / "clickhouse"
        mock_file_path = MagicMock()
        mock_resolved_path = MagicMock()
        mock_parent_path = MagicMock()
        mock_migrations_path = MagicMock()
        mock_clickhouse_path = MagicMock()

        MockPath.return_value = mock_file_path
        mock_file_path.resolve.return_value = mock_resolved_path
        mock_resolved_path.parent = mock_parent_path
        mock_parent_path.__truediv__.return_value = mock_migrations_path
        mock_migrations_path.__truediv__.return_value = mock_clickhouse_path
        
        mock_clickhouse_path.exists.return_value = True
        mock_clickhouse_path.glob.return_value = [mock_sql_file]

        store = ClickHouseStore("clickhouse://localhost:8123/default")
        async with store as s:
            assert s.client is mock_client

    get_client.assert_called_once_with(dsn="clickhouse://localhost:8123/default")
    # Expect 2 calls because our mock SQL has 2 statements
    assert mock_client.command.call_count == 2
    mock_client.close.assert_called_once()


@pytest.mark.asyncio
async def test_clickhouse_store_insert_git_file_data_calls_insert():
    test_repo_id = uuid.uuid4()
    file_data = [
        GitFile(
            repo_id=test_repo_id,
            path="file.txt",
            executable=False,
            contents="content",
        )
    ]

    mock_client = MagicMock()
    mock_client.command = MagicMock()
    mock_client.insert = MagicMock()
    mock_client.query = MagicMock(return_value=MagicMock(result_rows=[]))
    mock_client.close = MagicMock()

    import sys
    from types import SimpleNamespace

    get_client = MagicMock(return_value=mock_client)
    fake_clickhouse_connect = SimpleNamespace(get_client=get_client)

    with patch.dict(sys.modules, {"clickhouse_connect": fake_clickhouse_connect}):
        store = ClickHouseStore("clickhouse://localhost:8123/default")
        async with store:
            await store.insert_git_file_data(file_data)

    args, kwargs = mock_client.insert.call_args
    assert args[0] == "git_files"
    assert kwargs["column_names"] == [
        "repo_id",
        "path",
        "executable",
        "contents",
        "_mergestat_synced_at",
    ]


# MongoDB Tests


def _create_mock_collection():
    """Helper function to create a mock MongoDB collection with standard methods."""
    mock_collection = MagicMock()
    # Mock async methods
    mock_collection.update_one = AsyncMock(return_value=MagicMock(upserted_id=None))
    mock_collection.bulk_write = AsyncMock(return_value=MagicMock())
    mock_collection.find_one = AsyncMock(return_value=None)
    mock_collection.find = MagicMock(
        return_value=MagicMock(to_list=AsyncMock(return_value=[]))
    )
    mock_collection.count_documents = AsyncMock(return_value=0)
    return mock_collection


@pytest_asyncio.fixture
async def mongo_store():
    """Create a MongoStore instance with mocked MongoDB client for testing."""
    # Mock the AsyncIOMotorClient to avoid actual connection
    with patch("storage.AsyncIOMotorClient") as mock_client_class:
        # Setup the mock client and database
        mock_client = MagicMock()
        mock_db = MagicMock()
        mock_collections = {}

        def get_collection(self_arg, name):
            if name not in mock_collections:
                mock_collections[name] = _create_mock_collection()
            return mock_collections[name]

        mock_db.__getitem__ = get_collection
        mock_db.name = "test_db"
        mock_client.__getitem__ = MagicMock(return_value=mock_db)
        mock_client.close = MagicMock()
        mock_client_class.return_value = mock_client

        # Create the store instance normally to test constructor
        store = MongoStore("mongodb://localhost:27017", db_name="test_db")

        # Manually set the db since we're not using the context manager
        store.db = mock_db

        yield store


@pytest.mark.asyncio
async def test_mongo_store_init_with_empty_connection_string():
    """Test that MongoStore raises error with empty connection string."""
    with pytest.raises(ValueError, match="MongoDB connection string is required"):
        MongoStore("")


@pytest.mark.asyncio
async def test_mongo_store_init_with_connection_string():
    """Test that MongoStore initializes properly with a connection string."""
    with patch("storage.AsyncIOMotorClient"):
        store = MongoStore("mongodb://localhost:27017/test_db")
        assert store.db_name is None
        assert store.db is None


@pytest.mark.asyncio
async def test_mongo_store_init_with_db_name():
    """Test that MongoStore initializes properly with explicit db_name."""
    with patch("storage.AsyncIOMotorClient"):
        store = MongoStore("mongodb://localhost:27017", db_name="my_database")
        assert store.db_name == "my_database"
        assert store.db is None


@pytest.mark.asyncio
async def test_mongo_store_context_manager_with_db_name():
    """Test that MongoStore can be used as an async context manager with explicit db_name."""
    with patch("storage.AsyncIOMotorClient") as mock_client_class:
        mock_client = MagicMock()
        mock_db = MagicMock()
        mock_db.name = "my_database"
        mock_client.__getitem__ = MagicMock(return_value=mock_db)
        mock_client_class.return_value = mock_client

        store = MongoStore("mongodb://localhost:27017", db_name="my_database")

        async with store as s:
            assert s.db is not None
            assert s == store
            mock_client.__getitem__.assert_called_once_with("my_database")


@pytest.mark.asyncio
async def test_mongo_store_context_manager_with_connection_string_db():
    """Test MongoStore context manager with database in connection string."""
    with patch("storage.AsyncIOMotorClient") as mock_client_class:
        mock_client = MagicMock()
        mock_db = MagicMock()
        mock_db.name = "mydb"
        mock_client.get_default_database = MagicMock(return_value=mock_db)
        mock_client_class.return_value = mock_client

        store = MongoStore("mongodb://localhost:27017/mydb")

        async with store as s:
            assert s.db is not None
            assert s == store


@pytest.mark.asyncio
async def test_mongo_store_context_manager_without_db_raises_error():
    """Test that MongoStore raises error when no database is specified."""
    from pymongo.errors import ConfigurationError

    with patch("storage.AsyncIOMotorClient") as mock_client_class:
        mock_client = MagicMock()
        mock_client.get_default_database = MagicMock(
            side_effect=ConfigurationError("No default database")
        )
        mock_client_class.return_value = mock_client

        store = MongoStore("mongodb://localhost:27017")

        with pytest.raises(ValueError, match="No default database specified"):
            async with store:
                pass


@pytest.mark.asyncio
async def test_mongo_store_insert_repo(mongo_store):
    """Test inserting a repository into MongoDB."""
    test_repo = Repo(
        id=uuid.uuid4(),
        repo="https://github.com/test/repo.git",
        ref="main",
        settings={},
        tags=[],
    )

    await mongo_store.insert_repo(test_repo)

    # Verify the repo was inserted (update_one was called)
    mongo_store.db["repos"].update_one.assert_called_once()
    call_args = mongo_store.db["repos"].update_one.call_args

    # Verify the filter includes the repo id
    assert call_args[0][0]["_id"] == str(test_repo.id)
    # Verify upsert is True
    assert call_args[1]["upsert"] is True


@pytest.mark.asyncio
async def test_mongo_store_insert_repo_upsert(mongo_store):
    """Test that inserting a duplicate repo updates instead of creating duplicate."""
    test_repo = Repo(
        id=uuid.uuid4(),
        repo="https://github.com/test/repo.git",
        ref="main",
        settings={},
        tags=[],
    )

    # Insert the repo twice with different refs
    await mongo_store.insert_repo(test_repo)
    test_repo.ref = "develop"
    await mongo_store.insert_repo(test_repo)

    # Verify update_one was called twice
    assert mongo_store.db["repos"].update_one.call_count == 2


@pytest.mark.asyncio
async def test_mongo_store_insert_git_file_data(mongo_store):
    """Test inserting git file data into MongoDB."""
    test_repo_id = uuid.uuid4()

    file_data = [
        GitFile(
            repo_id=test_repo_id,
            path="file1.txt",
            executable=False,
            contents="content1",
        ),
        GitFile(
            repo_id=test_repo_id,
            path="file2.txt",
            executable=True,
            contents="content2",
        ),
    ]

    await mongo_store.insert_git_file_data(file_data)

    # Verify bulk_write was called
    mongo_store.db["git_files"].bulk_write.assert_called_once()
    call_args = mongo_store.db["git_files"].bulk_write.call_args

    # Verify operations list has 2 items
    operations = call_args[0][0]
    assert len(operations) == 2
    # Verify ordered=False
    assert call_args[1]["ordered"] is False


@pytest.mark.asyncio
async def test_mongo_store_insert_git_file_data_empty_list(mongo_store):
    """Test that inserting an empty list does not cause an error."""
    await mongo_store.insert_git_file_data([])

    # Verify bulk_write was not called
    mongo_store.db["git_files"].bulk_write.assert_not_called()


@pytest.mark.asyncio
async def test_mongo_store_insert_git_file_data_upsert(mongo_store):
    """Test that inserting duplicate file data updates instead of creating duplicates."""
    test_repo_id = uuid.uuid4()

    file_data = [
        GitFile(
            repo_id=test_repo_id,
            path="file.txt",
            executable=False,
            contents="original content",
        ),
    ]

    await mongo_store.insert_git_file_data(file_data)

    # Update the same file with new content
    file_data[0].contents = "updated content"
    await mongo_store.insert_git_file_data(file_data)

    # Verify bulk_write was called twice (upsert behavior)
    assert mongo_store.db["git_files"].bulk_write.call_count == 2


@pytest.mark.asyncio
async def test_mongo_store_insert_git_commit_data(mongo_store):
    """Test inserting git commit data into MongoDB."""
    test_repo_id = uuid.uuid4()

    commit_data = [
        GitCommit(
            repo_id=test_repo_id,
            hash="abc123",
            message="Initial commit",
            author_name="Test Author",
            author_email="test@example.com",
            author_when=datetime(2024, 1, 1, tzinfo=timezone.utc),
            committer_name="Test Committer",
            committer_email="committer@example.com",
            committer_when=datetime(2024, 1, 1, tzinfo=timezone.utc),
            parents=0,
        ),
        GitCommit(
            repo_id=test_repo_id,
            hash="def456",
            message="Second commit",
            author_name="Test Author",
            author_email="test@example.com",
            author_when=datetime(2024, 1, 2, tzinfo=timezone.utc),
            committer_name="Test Committer",
            committer_email="committer@example.com",
            committer_when=datetime(2024, 1, 2, tzinfo=timezone.utc),
            parents=1,
        ),
    ]

    await mongo_store.insert_git_commit_data(commit_data)

    # Verify bulk_write was called
    mongo_store.db["git_commits"].bulk_write.assert_called_once()
    call_args = mongo_store.db["git_commits"].bulk_write.call_args

    # Verify operations list has 2 items
    operations = call_args[0][0]
    assert len(operations) == 2


@pytest.mark.asyncio
async def test_mongo_store_insert_git_commit_data_empty_list(mongo_store):
    """Test that inserting an empty list does not cause an error."""
    await mongo_store.insert_git_commit_data([])

    # Verify bulk_write was not called
    mongo_store.db["git_commits"].bulk_write.assert_not_called()


@pytest.mark.asyncio
async def test_mongo_store_insert_git_commit_stats(mongo_store):
    """Test inserting git commit stats into MongoDB."""
    test_repo_id = uuid.uuid4()

    commit_stats = [
        GitCommitStat(
            repo_id=test_repo_id,
            commit_hash="abc123",
            file_path="file1.txt",
            additions=10,
            deletions=5,
            old_file_mode="100644",
            new_file_mode="100644",
        ),
        GitCommitStat(
            repo_id=test_repo_id,
            commit_hash="abc123",
            file_path="file2.txt",
            additions=20,
            deletions=3,
            old_file_mode="100644",
            new_file_mode="100755",
        ),
    ]

    await mongo_store.insert_git_commit_stats(commit_stats)

    # Verify bulk_write was called
    mongo_store.db["git_commit_stats"].bulk_write.assert_called_once()
    call_args = mongo_store.db["git_commit_stats"].bulk_write.call_args

    # Verify operations list has 2 items
    operations = call_args[0][0]
    assert len(operations) == 2


@pytest.mark.asyncio
async def test_mongo_store_insert_git_commit_stats_empty_list(mongo_store):
    """Test that inserting an empty list does not cause an error."""
    await mongo_store.insert_git_commit_stats([])

    # Verify bulk_write was not called
    mongo_store.db["git_commit_stats"].bulk_write.assert_not_called()


@pytest.mark.asyncio
async def test_mongo_store_insert_blame_data(mongo_store):
    """Test inserting git blame data into MongoDB."""
    test_repo_id = uuid.uuid4()

    blame_data = [
        GitBlame(
            repo_id=test_repo_id,
            path="file.txt",
            line_no=1,
            author_email="author@example.com",
            author_name="Test Author",
            author_when=datetime(2024, 1, 1, tzinfo=timezone.utc),
            commit_hash="abc123",
            line="line 1 content",
        ),
        GitBlame(
            repo_id=test_repo_id,
            path="file.txt",
            line_no=2,
            author_email="author@example.com",
            author_name="Test Author",
            author_when=datetime(2024, 1, 1, tzinfo=timezone.utc),
            commit_hash="abc123",
            line="line 2 content",
        ),
    ]

    await mongo_store.insert_blame_data(blame_data)

    # Verify bulk_write was called
    mongo_store.db["git_blame"].bulk_write.assert_called_once()
    call_args = mongo_store.db["git_blame"].bulk_write.call_args

    # Verify operations list has 2 items
    operations = call_args[0][0]
    assert len(operations) == 2


@pytest.mark.asyncio
async def test_mongo_store_insert_blame_data_empty_list(mongo_store):
    """Test that inserting an empty list does not cause an error."""
    await mongo_store.insert_blame_data([])

    # Verify bulk_write was not called
    mongo_store.db["git_blame"].bulk_write.assert_not_called()


@pytest.mark.asyncio
async def test_mongo_store_upsert_many_with_dict_payload(mongo_store):
    """Test _upsert_many with dict payload instead of model instances."""
    test_data = [
        {"repo_id": "test-repo", "path": "file1.txt", "contents": "content1"},
        {"repo_id": "test-repo", "path": "file2.txt", "contents": "content2"},
    ]

    await mongo_store._upsert_many(
        "test_collection",
        test_data,
        lambda obj: f"{obj['repo_id']}:{obj['path']}",
    )

    # Verify bulk_write was called on test_collection
    mongo_store.db["test_collection"].bulk_write.assert_called_once()
    call_args = mongo_store.db["test_collection"].bulk_write.call_args

    # Verify operations list has 2 items
    operations = call_args[0][0]
    assert len(operations) == 2


@pytest.mark.asyncio
async def test_mongo_store_bulk_operations_unordered(mongo_store):
    """Test that bulk operations are performed unordered (continue on error)."""
    test_repo_id = uuid.uuid4()

    # Create a large batch to test bulk operations
    file_data = [
        GitFile(
            repo_id=test_repo_id,
            path=f"file{i}.txt",
            executable=False,
            contents=f"content{i}",
        )
        for i in range(100)
    ]

    await mongo_store.insert_git_file_data(file_data)

    # Verify bulk_write was called with ordered=False
    mongo_store.db["git_files"].bulk_write.assert_called_once()
    call_args = mongo_store.db["git_files"].bulk_write.call_args
    assert call_args[1]["ordered"] is False

    # Verify 100 operations
    operations = call_args[0][0]
    assert len(operations) == 100


@pytest.mark.asyncio
async def test_mongo_store_connection_cleanup():
    """Test that MongoStore properly closes connection on exit."""
    with patch("storage.AsyncIOMotorClient") as mock_client_class:
        mock_client = MagicMock()
        mock_db = MagicMock()
        mock_client.__getitem__ = MagicMock(return_value=mock_db)
        mock_client.close = MagicMock()
        mock_client_class.return_value = mock_client

        store = MongoStore("mongodb://localhost:27017", db_name="test_db")

        async with store:
            assert store.db is not None

        # Verify client.close() was called
        mock_client.close.assert_called_once()


@pytest.mark.asyncio
async def test_mongo_store_upsert_id_builder_functionality(mongo_store):
    """Test that _upsert_many correctly uses id_builder to create composite keys."""
    test_repo_id = uuid.uuid4()

    file_data = [
        GitFile(
            repo_id=test_repo_id,
            path="dir/file.txt",
            executable=False,
            contents="content",
        ),
    ]

    await mongo_store.insert_git_file_data(file_data)

    # Verify the id_builder created correct composite key by checking bulk_write call
    call_args = mongo_store.db["git_files"].bulk_write.call_args
    operations = call_args[0][0]

    # Verify that bulk_write was called with UpdateOne operations
    assert isinstance(operations[0], UpdateOne)
    # The UpdateOne should have been created with the composite key
    # We verify this indirectly by checking that bulk_write was called with one operation
    assert len(operations) == 1


class TestMongoStoreMultipleRepos:
    """Test MongoDB behavior across multiple repos and multiple runs.

    These tests simulate the real-world scenario where the same database
    is used across multiple sync runs for different repositories.
    """

    @pytest_asyncio.fixture
    async def persistent_mongo_store(self):
        """Create a MongoStore with persistent in-memory state across operations."""
        # In-memory storage to simulate MongoDB collections
        collections_data = {}

        def create_persistent_collection(name):
            """Create a mock collection that persists data in memory."""
            if name not in collections_data:
                collections_data[name] = {}

            collection = MagicMock()
            collection.name = name

            async def mock_update_one(filter_dict, update_dict, upsert=False):
                doc_id = filter_dict.get("_id")
                if doc_id is not None:
                    if upsert or doc_id in collections_data[name]:
                        collections_data[name][doc_id] = update_dict.get("$set", {})
                return MagicMock(
                    modified_count=1, upserted_id=doc_id if upsert else None
                )

            async def mock_bulk_write(operations, ordered=True):
                for op in operations:
                    # Extract the filter and update from UpdateOne
                    # UpdateOne stores _filter and _doc internally
                    doc_id = op._filter.get("_id")
                    doc_data = op._doc.get("$set", {})
                    if doc_id is not None:
                        collections_data[name][doc_id] = doc_data
                return MagicMock(modified_count=len(operations))

            async def mock_find_to_list(length=None):
                return list(collections_data[name].values())

            async def mock_count_documents(filter_dict):
                if not filter_dict:
                    return len(collections_data[name])
                # Simple filter matching
                count = 0
                for doc in collections_data[name].values():
                    match = True
                    for key, value in filter_dict.items():
                        if doc.get(key) != value:
                            match = False
                            break
                    if match:
                        count += 1
                return count

            collection.update_one = mock_update_one
            collection.bulk_write = mock_bulk_write
            collection.find = MagicMock(
                return_value=MagicMock(to_list=mock_find_to_list)
            )
            collection.count_documents = mock_count_documents

            # Store reference to raw data for test assertions
            collection._data = collections_data[name]

            return collection

        with patch("storage.AsyncIOMotorClient") as mock_client_class:
            mock_client = MagicMock()
            mock_db = MagicMock()
            mock_db.name = "test_db"

            created_collections = {}

            def get_collection(name):
                if name not in created_collections:
                    created_collections[name] = create_persistent_collection(name)
                return created_collections[name]

            mock_db.__getitem__ = lambda self, name: get_collection(name)
            mock_client.__getitem__ = MagicMock(return_value=mock_db)
            mock_client.close = MagicMock()
            mock_client_class.return_value = mock_client

            store = MongoStore("mongodb://localhost:27017", db_name="test_db")
            store.db = mock_db
            store._collections_data = collections_data  # Expose for assertions

            yield store

    @pytest.mark.asyncio
    async def test_commits_from_two_different_repos_both_inserted(
        self, persistent_mongo_store
    ):
        """Test that commits from two different repos are both stored."""
        repo1_id = uuid.uuid4()
        repo2_id = uuid.uuid4()

        # First "run" - insert commits from repo1
        commits_repo1 = [
            GitCommit(
                repo_id=repo1_id,
                hash="abc123",
                message="Commit from repo1",
                author_name="Author 1",
                author_email="author1@example.com",
                author_when=datetime(2024, 1, 1, tzinfo=timezone.utc),
                committer_name="Committer 1",
                committer_email="committer1@example.com",
                committer_when=datetime(2024, 1, 1, tzinfo=timezone.utc),
                parents=0,
            ),
        ]
        await persistent_mongo_store.insert_git_commit_data(commits_repo1)

        # Second "run" - insert commits from repo2
        commits_repo2 = [
            GitCommit(
                repo_id=repo2_id,
                hash="def456",
                message="Commit from repo2",
                author_name="Author 2",
                author_email="author2@example.com",
                author_when=datetime(2024, 1, 2, tzinfo=timezone.utc),
                committer_name="Committer 2",
                committer_email="committer2@example.com",
                committer_when=datetime(2024, 1, 2, tzinfo=timezone.utc),
                parents=0,
            ),
        ]
        await persistent_mongo_store.insert_git_commit_data(commits_repo2)

        # Verify both commits exist in the collection
        commits_collection = persistent_mongo_store.db["git_commits"]
        assert len(commits_collection._data) == 2

        # Verify both repo_ids are present
        repo_ids_in_db = {
            doc.get("repo_id") for doc in commits_collection._data.values()
        }
        assert str(repo1_id) in repo_ids_in_db
        assert str(repo2_id) in repo_ids_in_db

    @pytest.mark.asyncio
    async def test_same_repo_multiple_runs_upserts_not_duplicates(
        self, persistent_mongo_store
    ):
        """Test that running the same repo twice updates existing commits, not duplicates."""
        repo_id = uuid.uuid4()

        # First run
        commits = [
            GitCommit(
                repo_id=repo_id,
                hash="abc123",
                message="Original message",
                author_name="Author",
                author_email="author@example.com",
                author_when=datetime(2024, 1, 1, tzinfo=timezone.utc),
                committer_name="Committer",
                committer_email="committer@example.com",
                committer_when=datetime(2024, 1, 1, tzinfo=timezone.utc),
                parents=0,
            ),
        ]
        await persistent_mongo_store.insert_git_commit_data(commits)

        # Second run with same commit hash (simulating re-sync)
        commits[0].message = "Updated message"
        await persistent_mongo_store.insert_git_commit_data(commits)

        # Should still only have 1 commit (upserted, not duplicated)
        commits_collection = persistent_mongo_store.db["git_commits"]
        assert len(commits_collection._data) == 1

        # Message should be updated
        stored_commit = list(commits_collection._data.values())[0]
        assert stored_commit["message"] == "Updated message"

    @pytest.mark.asyncio
    async def test_composite_key_differentiates_repos_with_same_commit_hash(
        self, persistent_mongo_store
    ):
        """Test that the same commit hash in different repos creates separate entries."""
        repo1_id = uuid.uuid4()
        repo2_id = uuid.uuid4()
        same_hash = "abc123"  # Same hash in both repos (e.g., copied/forked repo)

        commit1 = GitCommit(
            repo_id=repo1_id,
            hash=same_hash,
            message="Commit in repo1",
            author_name="Author",
            author_email="author@example.com",
            author_when=datetime(2024, 1, 1, tzinfo=timezone.utc),
            committer_name="Committer",
            committer_email="committer@example.com",
            committer_when=datetime(2024, 1, 1, tzinfo=timezone.utc),
            parents=0,
        )
        commit2 = GitCommit(
            repo_id=repo2_id,
            hash=same_hash,
            message="Same hash but different repo",
            author_name="Author",
            author_email="author@example.com",
            author_when=datetime(2024, 1, 1, tzinfo=timezone.utc),
            committer_name="Committer",
            committer_email="committer@example.com",
            committer_when=datetime(2024, 1, 1, tzinfo=timezone.utc),
            parents=0,
        )

        await persistent_mongo_store.insert_git_commit_data([commit1])
        await persistent_mongo_store.insert_git_commit_data([commit2])

        # Should have 2 separate entries because composite key includes repo_id
        commits_collection = persistent_mongo_store.db["git_commits"]
        assert len(commits_collection._data) == 2

    @pytest.mark.asyncio
    async def test_new_repo_commits_added_to_existing_data(
        self, persistent_mongo_store
    ):
        """Test that adding a new repo doesn't clear existing repos' data."""
        repo1_id = uuid.uuid4()
        repo2_id = uuid.uuid4()

        # Insert repo1 with multiple commits
        commits_repo1 = [
            GitCommit(
                repo_id=repo1_id,
                hash=f"hash{i}",
                message=f"Commit {i}",
                author_name="Author",
                author_email="author@example.com",
                author_when=datetime(2024, 1, i + 1, tzinfo=timezone.utc),
                committer_name="Committer",
                committer_email="committer@example.com",
                committer_when=datetime(2024, 1, i + 1, tzinfo=timezone.utc),
                parents=0,
            )
            for i in range(5)
        ]
        await persistent_mongo_store.insert_git_commit_data(commits_repo1)

        # Verify 5 commits for repo1
        commits_collection = persistent_mongo_store.db["git_commits"]
        assert len(commits_collection._data) == 5

        # Insert repo2 with 3 commits
        commits_repo2 = [
            GitCommit(
                repo_id=repo2_id,
                hash=f"repo2hash{i}",
                message=f"Repo2 Commit {i}",
                author_name="Author 2",
                author_email="author2@example.com",
                author_when=datetime(2024, 2, i + 1, tzinfo=timezone.utc),
                committer_name="Committer 2",
                committer_email="committer2@example.com",
                committer_when=datetime(2024, 2, i + 1, tzinfo=timezone.utc),
                parents=0,
            )
            for i in range(3)
        ]
        await persistent_mongo_store.insert_git_commit_data(commits_repo2)

        # Should now have 8 total commits (5 from repo1 + 3 from repo2)
        assert len(commits_collection._data) == 8

        # Verify repo1 commits still exist
        repo1_commits = [
            doc
            for doc in commits_collection._data.values()
            if doc.get("repo_id") == str(repo1_id)
        ]
        assert len(repo1_commits) == 5

        # Verify repo2 commits exist
        repo2_commits = [
            doc
            for doc in commits_collection._data.values()
            if doc.get("repo_id") == str(repo2_id)
        ]
        assert len(repo2_commits) == 3


# SQLite-specific Tests


class TestSQLiteAsyncStorage:
    """Test SQLite async storage specific functionality."""

    @pytest.fixture
    def sqlite_memory_url(self):
        """Return a SQLite in-memory database URL for testing."""
        return "sqlite+aiosqlite:///:memory:"

    @pytest.fixture
    def sqlite_file_url(self, tmp_path):
        """Return a SQLite file-based database URL for testing."""
        db_path = tmp_path / "test_mergestat.db"
        return f"sqlite+aiosqlite:///{db_path}"

    @pytest_asyncio.fixture
    async def sqlite_memory_store(self, sqlite_memory_url):
        """Create a SQLAlchemyStore instance with an in-memory SQLite database."""
        store = SQLAlchemyStore(sqlite_memory_url)

        # Create tables
        async with store.engine.begin() as conn:
            await conn.run_sync(Base.metadata.create_all)

        yield store

        # Cleanup
        async with store.engine.begin() as conn:
            await conn.run_sync(Base.metadata.drop_all)
        await store.engine.dispose()

    @pytest_asyncio.fixture
    async def sqlite_file_store(self, sqlite_file_url):
        """Create a SQLAlchemyStore instance with a file-based SQLite database."""
        store = SQLAlchemyStore(sqlite_file_url)

        # Create tables
        async with store.engine.begin() as conn:
            await conn.run_sync(Base.metadata.create_all)

        yield store

        # Cleanup
        async with store.engine.begin() as conn:
            await conn.run_sync(Base.metadata.drop_all)
        await store.engine.dispose()

    @pytest.mark.asyncio
    async def test_sqlite_memory_store_connection_string_detection(
        self, sqlite_memory_url
    ):
        """Test that SQLite connection strings are correctly identified."""
        store = SQLAlchemyStore(sqlite_memory_url)

        # SQLite should not have pool_size configured (uses StaticPool by default)
        engine = store.engine
        assert "sqlite" in str(engine.url).lower()
        await store.engine.dispose()

    @pytest.mark.asyncio
    async def test_sqlite_file_store_connection_string_detection(self, sqlite_file_url):
        """Test that SQLite file-based connection strings are correctly identified."""
        store = SQLAlchemyStore(sqlite_file_url)

        engine = store.engine
        assert "sqlite" in str(engine.url).lower()
        await store.engine.dispose()

    @pytest.mark.asyncio
    async def test_sqlite_memory_store_insert_and_query(self, sqlite_memory_store):
        """Test inserting and querying data with SQLite in-memory database."""
        test_repo = Repo(
            id=uuid.uuid4(),
            repo="https://github.com/test/sqlite-test.git",
            ref="main",
            settings={},
            tags=[],
        )

        async with sqlite_memory_store as store:
            await store.insert_repo(test_repo)

            # Verify the repo was inserted
            result = await store.session.execute(
                select(Repo).where(Repo.id == test_repo.id)
            )
            saved_repo = result.scalar_one_or_none()

            assert saved_repo is not None
            assert saved_repo.id == test_repo.id
            assert saved_repo.repo == "https://github.com/test/sqlite-test.git"

    @pytest.mark.asyncio
    async def test_sqlite_file_store_insert_and_query(self, sqlite_file_store):
        """Test inserting and querying data with SQLite file database."""
        test_repo = Repo(
            id=uuid.uuid4(),
            repo="https://github.com/test/sqlite-file-test.git",
            ref="develop",
            settings={"source": "test"},
            tags=["test"],
        )

        async with sqlite_file_store as store:
            await store.insert_repo(test_repo)

            # Verify the repo was inserted
            result = await store.session.execute(
                select(Repo).where(Repo.id == test_repo.id)
            )
            saved_repo = result.scalar_one_or_none()

            assert saved_repo is not None
            assert saved_repo.id == test_repo.id
            assert saved_repo.repo == "https://github.com/test/sqlite-file-test.git"

    @pytest.mark.asyncio
    async def test_sqlite_store_insert_all_data_types(self, sqlite_memory_store):
        """Test inserting all data types (commits, files, stats, blame) into SQLite."""
        test_repo_id = uuid.uuid4()
        test_repo = Repo(
            id=test_repo_id,
            repo="https://github.com/test/all-types.git",
            ref="main",
            settings={},
            tags=[],
        )

        file_data = [
            GitFile(
                repo_id=test_repo_id,
                path="src/main.py",
                executable=False,
                contents="print('hello')",
            ),
        ]

        commit_data = [
            GitCommit(
                repo_id=test_repo_id,
                hash="abc123def456",
                message="Initial commit",
                author_name="Test Author",
                author_email="test@example.com",
                author_when=datetime(2024, 1, 1, tzinfo=timezone.utc),
                committer_name="Test Committer",
                committer_email="committer@example.com",
                committer_when=datetime(2024, 1, 1, tzinfo=timezone.utc),
                parents=0,
            ),
        ]

        commit_stats = [
            GitCommitStat(
                repo_id=test_repo_id,
                commit_hash="abc123def456",
                file_path="src/main.py",
                additions=10,
                deletions=0,
                old_file_mode="100644",
                new_file_mode="100644",
            ),
        ]

        blame_data = [
            GitBlame(
                repo_id=test_repo_id,
                path="src/main.py",
                line_no=1,
                author_email="test@example.com",
                author_name="Test Author",
                author_when=datetime(2024, 1, 1, tzinfo=timezone.utc),
                commit_hash="abc123def456",
                line="print('hello')",
            ),
        ]

        async with sqlite_memory_store as store:
            # Insert all data types
            await store.insert_repo(test_repo)
            await store.insert_git_file_data(file_data)
            await store.insert_git_commit_data(commit_data)
            await store.insert_git_commit_stats(commit_stats)
            await store.insert_blame_data(blame_data)

            # Verify all data was inserted
            repo_result = await store.session.execute(
                select(Repo).where(Repo.id == test_repo_id)
            )
            assert repo_result.scalar_one_or_none() is not None

            file_result = await store.session.execute(
                select(GitFile).where(GitFile.repo_id == test_repo_id)
            )
            assert len(file_result.scalars().all()) == 1

            commit_result = await store.session.execute(
                select(GitCommit).where(GitCommit.repo_id == test_repo_id)
            )
            assert len(commit_result.scalars().all()) == 1

            stats_result = await store.session.execute(
                select(GitCommitStat).where(GitCommitStat.repo_id == test_repo_id)
            )
            assert len(stats_result.scalars().all()) == 1

            blame_result = await store.session.execute(
                select(GitBlame).where(GitBlame.repo_id == test_repo_id)
            )
            assert len(blame_result.scalars().all()) == 1

    @pytest.mark.asyncio
    async def test_sqlite_store_no_connection_pooling_params(self, sqlite_memory_url):
        """Test that SQLite stores don't use PostgreSQL-specific pooling params."""
        store = SQLAlchemyStore(sqlite_memory_url)

        # The engine should be created without pool_size, max_overflow params
        # which are PostgreSQL-specific
        engine = store.engine

        # Verify the engine was created successfully
        assert engine is not None

        # SQLite connections should work
        async with engine.begin() as conn:
            await conn.run_sync(Base.metadata.create_all)

        await store.engine.dispose()

    @pytest.mark.asyncio
    async def test_sqlite_store_multiple_sessions(self, sqlite_file_url):
        """Test that multiple sessions can be created with SQLite file store."""
        store = SQLAlchemyStore(sqlite_file_url)

        # Create tables
        async with store.engine.begin() as conn:
            await conn.run_sync(Base.metadata.create_all)

        test_repo = Repo(
            id=uuid.uuid4(),
            repo="https://github.com/test/multi-session.git",
            ref="main",
            settings={},
            tags=[],
        )

        # First session - insert data
        async with store as s1:
            await s1.insert_repo(test_repo)

        # Second session - verify data persists
        async with store as s2:
            result = await s2.session.execute(
                select(Repo).where(Repo.id == test_repo.id)
            )
            saved_repo = result.scalar_one_or_none()
            assert saved_repo is not None
            assert saved_repo.id == test_repo.id

        # Cleanup
        async with store.engine.begin() as conn:
            await conn.run_sync(Base.metadata.drop_all)
        await store.engine.dispose()

    @pytest.mark.asyncio
    async def test_sqlite_store_batch_inserts(self, sqlite_memory_store):
        """Test that batch inserts work correctly with SQLite."""
        test_repo_id = uuid.uuid4()
        test_repo = Repo(
            id=test_repo_id,
            repo="https://github.com/test/batch.git",
            ref="main",
            settings={},
            tags=[],
        )

        async with sqlite_memory_store as store:
            await store.insert_repo(test_repo)

            # Create multiple files to insert in batches
            files = [
                GitFile(
                    repo_id=test_repo_id,
                    path=f"file{i}.txt",
                    executable=False,
                    contents=f"content{i}",
                )
                for i in range(10)
            ]

            # Insert in batches
            await store.insert_git_file_data(files[:5])
            await store.insert_git_file_data(files[5:])

            # Verify all files were inserted
            result = await store.session.execute(
                select(GitFile).where(GitFile.repo_id == test_repo_id)
            )
            saved_files = result.scalars().all()
            assert len(saved_files) == 10



"""
Exception types for connector operations.
"""


class ConnectorException(Exception):
    """Base exception for all connector errors."""


class RateLimitException(ConnectorException):
    """Raised when API rate limit is exceeded."""

    def __init__(
        self,
        message: str = "API rate limit exceeded",
        retry_after_seconds: float | None = None,
    ) -> None:
        super().__init__(message)
        self.retry_after_seconds = retry_after_seconds


class AuthenticationException(ConnectorException):
    """Raised when authentication fails."""


class NotFoundException(ConnectorException):
    """Raised when a resource is not found."""


class PaginationException(ConnectorException):
    """Raised when pagination fails."""


class APIException(ConnectorException):
    """Raised when API returns an error."""



"""Shared test fixtures for the test suite."""

import mimetypes
import os
import uuid
from pathlib import Path

import pytest
from git import Repo as GitRepo


@pytest.fixture
def repo_path():
    """Return the path to the current repository for testing."""
    return str(Path(__file__).parent.parent)


@pytest.fixture
def repo_uuid():
    """Return a test UUID for the repository."""
    return uuid.uuid4()


@pytest.fixture
def git_repo(repo_path):
    """Return a GitRepo instance for testing."""
    return GitRepo(repo_path)


@pytest.fixture
def test_file(repo_path):
    """Return a path to an existing file in the repository."""
    return os.path.join(repo_path, "README.md")


def pytest_configure(config):
    # Ensure TypeScript files are treated as text, not video/mp2t.
    mimetypes.add_type("text/x-typescript", ".ts")



from __future__ import annotations

import uuid
from datetime import date, datetime, time, timedelta, timezone
from typing import Dict, List, Sequence, Tuple

from metrics.schemas import IncidentMetricsDailyRecord, IncidentRow


def _utc_day_window(day: date) -> Tuple[datetime, datetime]:
    start = datetime.combine(day, time.min, tzinfo=timezone.utc)
    end = start + timedelta(days=1)
    return start, end


def _to_utc(dt: datetime) -> datetime:
    if dt.tzinfo is None:
        return dt.replace(tzinfo=timezone.utc)
    return dt.astimezone(timezone.utc)


def _percentile(values: Sequence[float], percentile: float) -> float:
    if not values:
        return 0.0
    if percentile <= 0:
        return float(min(values))
    if percentile >= 100:
        return float(max(values))
    sorted_vals = sorted(float(v) for v in values)
    if len(sorted_vals) == 1:
        return float(sorted_vals[0])
    rank = (len(sorted_vals) - 1) * (float(percentile) / 100.0)
    lo = int(rank)
    hi = min(lo + 1, len(sorted_vals) - 1)
    frac = rank - lo
    return sorted_vals[lo] * (1.0 - frac) + sorted_vals[hi] * frac


def compute_incident_metrics_daily(
    *,
    day: date,
    incidents: Sequence[IncidentRow],
    computed_at: datetime,
) -> List[IncidentMetricsDailyRecord]:
    """
    Compute MTTR distributions for incidents resolved on the given day.
    """
    start, end = _utc_day_window(day)
    computed_at_utc = _to_utc(computed_at)

    by_repo: Dict[str, Dict[str, object]] = {}
    for row in incidents:
        resolved_at = row.get("resolved_at")
        if not isinstance(resolved_at, datetime):
            continue
        resolved_at = _to_utc(resolved_at)
        if not (start <= resolved_at < end):
            continue

        repo_id = str(row["repo_id"])
        bucket = by_repo.get(repo_id)
        if bucket is None:
            bucket = {"incidents": 0, "mttr_hours": []}
            by_repo[repo_id] = bucket

        bucket["incidents"] = int(bucket["incidents"]) + 1
        started_at = _to_utc(row["started_at"])
        mttr = (resolved_at - started_at).total_seconds() / 3600.0
        if mttr >= 0:
            bucket["mttr_hours"].append(float(mttr))

    records: List[IncidentMetricsDailyRecord] = []
    for repo_id, bucket in sorted(by_repo.items(), key=lambda kv: kv[0]):
        mttr_hours: List[float] = list(bucket["mttr_hours"])
        records.append(
            IncidentMetricsDailyRecord(
                repo_id=uuid.UUID(repo_id),
                day=day,
                incidents_count=int(bucket["incidents"]),
                mttr_p50_hours=float(_percentile(mttr_hours, 50.0))
                if mttr_hours
                else None,
                mttr_p90_hours=float(_percentile(mttr_hours, 90.0))
                if mttr_hours
                else None,
                computed_at=computed_at_utc,
            )
        )

    return records



import logging
from pathlib import Path
from typing import Dict

import yaml

logger = logging.getLogger(__name__)

class IssueTypeNormalizer:
    def __init__(self, config_path: Path):
        self.mapping = self._load_config(config_path)

    def _load_config(self, path: Path) -> Dict:
        if not path.exists():
            logger.warning(f"Issue type mapping config not found at {path}, using defaults")
            return {}
        with open(path, "r") as f:
            return yaml.safe_load(f)

    def normalize(self, provider: str, raw_type: str, labels: list[str]) -> str:
        provider = provider.lower()
        
        # Jira: direct mapping from raw_type
        if provider == "jira":
            mapping = self.mapping.get("jira", {})
            return mapping.get(raw_type, "unassigned")

        # GitHub: infer from labels
        if provider == "github":
            mapping = self.mapping.get("github_labels", {})
            # Look for matching labels
            for label in labels:
                norm = mapping.get(label.lower())
                if norm:
                    return norm
            # Fallback or default
            return "unassigned"

        # GitLab: infer from labels (scoped or simple)
        if provider == "gitlab":
            mapping = self.mapping.get("gitlab_labels", {})
            for label in labels:
                norm = mapping.get(label.lower())
                if norm:
                    return norm
            return "unassigned"

        return "unassigned"



import logging
import uuid
from datetime import date, datetime, timedelta, timezone
from pathlib import Path
from typing import Optional, Any

import git
from analytics.complexity import ComplexityScanner
from metrics.schemas import FileComplexitySnapshot, RepoComplexityDaily
from metrics.sinks.clickhouse import ClickHouseMetricsSink
from metrics.sinks.sqlite import SQLiteMetricsSink
from metrics.sinks.mongo import MongoMetricsSink
from storage import detect_db_type

logger = logging.getLogger(__name__)

REPO_ROOT = Path(__file__).resolve().parent.parent.parent


def _normalize_sqlite_url(db_url: str) -> str:
    if "sqlite+aiosqlite://" in db_url:
        return db_url.replace("sqlite+aiosqlite://", "sqlite://", 1)
    return db_url


def _normalize_postgres_url(db_url: str) -> str:
    if "postgresql+asyncpg://" in db_url:
        return db_url.replace("postgresql+asyncpg://", "postgresql://", 1)
    return db_url


def run_complexity_scan_job(
    *,
    repo_path: Path,
    repo_id: uuid.UUID,
    db_url: str,
    date: date,
    backfill_days: int = 1,
    ref: str = "HEAD",
    sink: Optional[Any] = None,
) -> None:
    if not db_url and not sink:
        raise ValueError("DB connection string or sink is required.")

    if not repo_path.exists():
        raise FileNotFoundError(f"Repo path {repo_path} does not exist.")

    repo = git.Repo(repo_path)
    config_path = REPO_ROOT / "config/complexity.yaml"
    scanner = ComplexityScanner(config_path=config_path)
    
    # Calculate date range
    start_date = date - timedelta(days=max(1, backfill_days) - 1)
    dates_to_process = [start_date + timedelta(days=i) for i in range(max(1, backfill_days))]

    own_sink = False
    if sink is None:
        backend = detect_db_type(db_url)
        own_sink = True
        if backend == "clickhouse":
            sink = ClickHouseMetricsSink(db_url)
        elif backend == "sqlite":
            sink = SQLiteMetricsSink(_normalize_sqlite_url(db_url))
        elif backend == "mongo":
            sink = MongoMetricsSink(db_url)
        elif backend == "postgres":
            sink = SQLiteMetricsSink(_normalize_postgres_url(db_url))
        else:
            raise ValueError(f"Unsupported backend for complexity job: {backend}")

    try:
        if own_sink:
            sink.ensure_tables()
        
        for d in dates_to_process:
            logger.info(f"Processing date: {d}")
            
            # Find commit for this day
            # "last commit before end of day"
            # git rev-list -1 --before="YYYY-MM-DD 23:59:59" ref
            end_of_day_ts = datetime.combine(d, datetime.max.time()).replace(tzinfo=timezone.utc).isoformat()
            
            try:
                commit_sha = repo.git.rev_list("-1", f"--before={end_of_day_ts}", ref)
            except git.Exc as e:
                logger.warning(f"Could not find commit for {d}: {e}")
                continue
                
            if not commit_sha:
                logger.warning(f"No commit found before {d} (ref={ref}). Skipping.")
                continue

            commit_sha = commit_sha.strip()
            logger.info(f"Scanning {repo_path} at commit {commit_sha} (as of {d})...")
            
            file_results = scanner.scan_git_ref(repo_path, commit_sha)
            logger.info(f"Scanned {len(file_results)} files.")

            if not file_results:
                logger.warning(f"No complexity data found for {d}.")
                continue

            # Aggregations
            computed_at = datetime.now(timezone.utc)
            
            snapshots = []
            total_loc = 0
            total_cc = 0
            total_high = 0
            total_very_high = 0
            
            for f in file_results:
                snapshots.append(FileComplexitySnapshot(
                    repo_id=repo_id,
                    as_of_day=d,
                    ref=commit_sha,
                    file_path=f.file_path,
                    language=f.language,
                    loc=f.loc,
                    functions_count=f.functions_count,
                    cyclomatic_total=f.cyclomatic_total,
                    cyclomatic_avg=f.cyclomatic_avg,
                    high_complexity_functions=f.high_complexity_functions,
                    very_high_complexity_functions=f.very_high_complexity_functions,
                    computed_at=computed_at
                ))
                
                total_loc += f.loc
                total_cc += f.cyclomatic_total
                total_high += f.high_complexity_functions
                total_very_high += f.very_high_complexity_functions

            cc_per_kloc = (total_cc / (total_loc / 1000.0)) if total_loc > 0 else 0.0
            
            repo_daily = RepoComplexityDaily(
                repo_id=repo_id,
                day=d,
                loc_total=total_loc,
                cyclomatic_total=total_cc,
                cyclomatic_per_kloc=cc_per_kloc,
                high_complexity_functions=total_high,
                very_high_complexity_functions=total_very_high,
                computed_at=computed_at
            )

            logger.info(f"Writing {len(snapshots)} file snapshots for {d}...")
            sink.write_file_complexity_snapshots(snapshots)
            logger.info(f"Writing repo daily summary for {d}...")
            sink.write_repo_complexity_daily([repo_daily])

    finally:
        if own_sink:
            sink.close()
    
    logger.info("Complexity job done.")



"""
Base connector class for Git repository connectors.

This module provides an abstract base class that defines the common interface
for all Git connectors (GitHub, GitLab, local, etc.).
"""

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Callable, List, Optional

from connectors.models import (
    Author,
    CommitStats,
    FileBlame,
    Organization,
    PullRequest,
    Repository,
    RepoStats,
)

logger = logging.getLogger(__name__)


@dataclass
class BatchResult:
    """Result of a batch repository processing operation."""

    repository: Repository
    stats: Optional[RepoStats] = None
    error: Optional[str] = None
    success: bool = True


class GitConnector(ABC):
    """
    Abstract base class for Git repository connectors.

    This class defines the common interface that all Git connectors
    (GitHub, GitLab, local) must implement.
    """

    def __init__(
        self,
        per_page: int = 100,
        max_workers: int = 4,
    ):
        """
        Initialize the base connector.

        :param per_page: Number of items per page for pagination.
        :param max_workers: Maximum concurrent workers for operations.
        """
        self.per_page = per_page
        self.max_workers = max_workers

    @abstractmethod
    def list_organizations(
        self,
        max_orgs: Optional[int] = None,
    ) -> List[Organization]:
        """
        List organizations/groups accessible to the authenticated user.

        :param max_orgs: Maximum number of organizations to retrieve.
        :return: List of Organization objects.
        """
        pass

    @abstractmethod
    def list_repositories(
        self,
        org_name: Optional[str] = None,
        user_name: Optional[str] = None,
        search: Optional[str] = None,
        pattern: Optional[str] = None,
        max_repos: Optional[int] = None,
    ) -> List[Repository]:
        """
        List repositories for an organization, user, or search query.

        :param org_name: Optional organization name.
        :param user_name: Optional user name.
        :param search: Optional search query.
        :param pattern: Optional fnmatch-style pattern to filter repositories.
        :param max_repos: Maximum number of repositories to retrieve.
        :return: List of Repository objects.
        """
        pass

    @abstractmethod
    def get_contributors(
        self,
        owner: str,
        repo: str,
        max_contributors: Optional[int] = None,
    ) -> List[Author]:
        """
        Get contributors for a repository.

        :param owner: Repository owner.
        :param repo: Repository name.
        :param max_contributors: Maximum number of contributors to retrieve.
        :return: List of Author objects.
        """
        pass

    @abstractmethod
    def get_commit_stats(
        self,
        owner: str,
        repo: str,
        sha: str,
    ) -> CommitStats:
        """
        Get statistics for a specific commit.

        :param owner: Repository owner.
        :param repo: Repository name.
        :param sha: Commit SHA.
        :return: CommitStats object.
        """
        pass

    @abstractmethod
    def get_repo_stats(
        self,
        owner: str,
        repo: str,
        max_commits: Optional[int] = None,
    ) -> RepoStats:
        """
        Get aggregated statistics for a repository.

        :param owner: Repository owner.
        :param repo: Repository name.
        :param max_commits: Maximum number of commits to analyze.
        :return: RepoStats object.
        """
        pass

    @abstractmethod
    def get_pull_requests(
        self,
        owner: str,
        repo: str,
        state: str = "all",
        max_prs: Optional[int] = None,
    ) -> List[PullRequest]:
        """
        Get pull requests/merge requests for a repository.

        :param owner: Repository owner.
        :param repo: Repository name.
        :param state: State filter ('open', 'closed', 'all').
        :param max_prs: Maximum number of pull requests to retrieve.
        :return: List of PullRequest objects.
        """
        pass

    @abstractmethod
    def get_file_blame(
        self,
        owner: str,
        repo: str,
        path: str,
        ref: str = "HEAD",
    ) -> FileBlame:
        """
        Get blame information for a file.

        :param owner: Repository owner.
        :param repo: Repository name.
        :param path: File path within the repository.
        :param ref: Git reference (branch, tag, or commit SHA).
        :return: FileBlame object.
        """
        pass

    @abstractmethod
    def get_repos_with_stats(
        self,
        org_name: Optional[str] = None,
        user_name: Optional[str] = None,
        pattern: Optional[str] = None,
        batch_size: int = 10,
        max_concurrent: int = 4,
        rate_limit_delay: float = 1.0,
        max_commits_per_repo: Optional[int] = None,
        max_repos: Optional[int] = None,
        on_repo_complete: Optional[Callable[[BatchResult], None]] = None,
    ) -> List[BatchResult]:
        """
        Get repositories and their stats with batch processing.

        Each connector must implement its own filtering and processing logic.

        :param org_name: Optional organization name.
        :param user_name: Optional user name.
        :param pattern: Optional fnmatch-style pattern to filter repos.
        :param batch_size: Number of repos to process in each batch.
        :param max_concurrent: Maximum concurrent workers for processing.
        :param rate_limit_delay: Delay in seconds between batches.
        :param max_commits_per_repo: Maximum commits to analyze per repository.
        :param max_repos: Maximum number of repositories to process.
        :param on_repo_complete: Callback function called after each repo.
        :return: List of BatchResult objects.
        """
        pass

    @abstractmethod
    async def get_repos_with_stats_async(
        self,
        org_name: Optional[str] = None,
        user_name: Optional[str] = None,
        pattern: Optional[str] = None,
        batch_size: int = 10,
        max_concurrent: int = 4,
        rate_limit_delay: float = 1.0,
        max_commits_per_repo: Optional[int] = None,
        max_repos: Optional[int] = None,
        on_repo_complete: Optional[Callable[[BatchResult], None]] = None,
    ) -> List[BatchResult]:
        """
        Async version of get_repos_with_stats.

        Each connector must implement its own async filtering and processing logic.

        :param org_name: Optional organization name.
        :param user_name: Optional user name.
        :param pattern: Optional fnmatch-style pattern to filter repos.
        :param batch_size: Number of repos to process in each batch.
        :param max_concurrent: Maximum concurrent workers for processing.
        :param rate_limit_delay: Delay in seconds between batches.
        :param max_commits_per_repo: Maximum commits to analyze per repository.
        :param max_repos: Maximum number of repositories to process.
        :param on_repo_complete: Callback function called after each repo.
        :return: List of BatchResult objects.
        """
        pass

    @abstractmethod
    def close(self) -> None:
        """Close the connector and cleanup resources."""
        pass

    def __enter__(self):
        """Context manager entry."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit."""
        self.close()
        return False



"""Tests for GitBlame functionality including repo instance reuse."""

import logging
import os
from unittest.mock import MagicMock, patch

from models.git import GitBlame, GitBlameMixin


class TestGitBlameMixin:
    """Test cases for GitBlameMixin.fetch_blame()."""

    def test_fetch_blame_without_repo_param(self, repo_path, test_file, repo_uuid):
        """Test fetch_blame creates its own Repo instance when none provided."""
        blame_data = GitBlameMixin.fetch_blame(repo_path, test_file, repo_uuid)

        # Should return a list of blame data tuples
        assert isinstance(blame_data, list)
        # README.md should have content
        if blame_data:
            # Each tuple should have 8 elements
            assert len(blame_data[0]) == 8
            # First element should be the repo_uuid
            assert blame_data[0][0] == repo_uuid

    def test_fetch_blame_with_repo_param(
        self, repo_path, test_file, repo_uuid, git_repo
    ):
        """Test fetch_blame reuses provided Repo instance."""
        blame_data = GitBlameMixin.fetch_blame(
            repo_path, test_file, repo_uuid, repo=git_repo
        )

        # Should return a list of blame data tuples
        assert isinstance(blame_data, list)
        if blame_data:
            # Each tuple should have 8 elements
            assert len(blame_data[0]) == 8
            # First element should be the repo_uuid
            assert blame_data[0][0] == repo_uuid

    def test_fetch_blame_repo_reuse_performance(self, repo_path, repo_uuid, git_repo):
        """Test that passing a repo instance avoids creating a new one."""
        # Mock the Repo class to track instantiation
        with patch("models.Repo") as mock_repo:
            # Call with existing repo - should not create a new one
            GitBlameMixin.fetch_blame(
                repo_path,
                os.path.join(repo_path, "README.md"),
                repo_uuid,
                repo=git_repo,
            )

            # Repo should not be called when we pass an existing instance
            mock_repo.assert_not_called()

    def test_fetch_blame_creates_repo_when_none_provided(self, repo_path, repo_uuid):
        """Test that fetch_blame creates a GitRepo when none is provided."""
        with patch("models.git.GitRepo") as mock_repo:
            mock_repo_instance = MagicMock()
            mock_repo_instance.blame.return_value = []
            mock_repo.return_value = mock_repo_instance

            GitBlameMixin.fetch_blame(
                repo_path, os.path.join(repo_path, "README.md"), repo_uuid, repo=None
            )

            # GitRepo should be instantiated when none is provided
            mock_repo.assert_called_once_with(repo_path)

    def test_fetch_blame_handles_errors_gracefully(
        self, repo_path, repo_uuid, git_repo, caplog
    ):
        """Test that fetch_blame handles errors and logs warning message."""
        with caplog.at_level(logging.WARNING):
            # Try to process a non-existent file
            blame_data = GitBlameMixin.fetch_blame(
                repo_path,
                os.path.join(repo_path, "nonexistent_file.xyz"),
                repo_uuid,
                repo=git_repo,
            )

            # Should return empty list on error
            assert blame_data == []

            # Should log warning message
            assert "Error processing" in caplog.text


class TestGitBlameProcessFile:
    """Test cases for GitBlame.process_file()."""

    def test_process_file_without_repo_param(self, repo_path, test_file, repo_uuid):
        """Test process_file works without a repo parameter."""
        blame_objects = GitBlame.process_file(repo_path, test_file, repo_uuid)

        assert isinstance(blame_objects, list)
        if blame_objects:
            # Should return GitBlame objects
            assert isinstance(blame_objects[0], GitBlame)
            # Check attributes are set
            assert blame_objects[0].repo_id == repo_uuid
            assert blame_objects[0].line_no >= 1
            assert blame_objects[0].path is not None

    def test_process_file_with_repo_param(
        self, repo_path, test_file, repo_uuid, git_repo
    ):
        """Test process_file reuses provided Repo instance."""
        blame_objects = GitBlame.process_file(
            repo_path, test_file, repo_uuid, repo=git_repo
        )

        assert isinstance(blame_objects, list)
        if blame_objects:
            assert isinstance(blame_objects[0], GitBlame)
            assert blame_objects[0].repo_id == repo_uuid

    def test_process_file_passes_repo_to_fetch_blame(
        self, repo_path, test_file, repo_uuid, git_repo
    ):
        """Test that process_file passes repo parameter to fetch_blame."""
        with patch.object(GitBlame, "fetch_blame") as mock_fetch:
            mock_fetch.return_value = []

            GitBlame.process_file(repo_path, test_file, repo_uuid, repo=git_repo)

            # Verify fetch_blame was called with the repo parameter
            mock_fetch.assert_called_once_with(
                repo_path, test_file, repo_uuid, repo=git_repo
            )

    def test_process_file_returns_correct_gitblame_objects(
        self, repo_path, test_file, repo_uuid, git_repo
    ):
        """Test that process_file correctly maps blame data to GitBlame objects."""
        # Create mock blame data
        mock_blame_data = [
            (
                repo_uuid,
                "test@example.com",
                "Test Author",
                "2024-01-01 00:00:00",
                "abc123",
                1,
                "line content",
                "README.md",
            )
        ]

        with patch.object(GitBlame, "fetch_blame", return_value=mock_blame_data):
            blame_objects = GitBlame.process_file(
                repo_path, test_file, repo_uuid, repo=git_repo
            )

            assert len(blame_objects) == 1
            obj = blame_objects[0]
            assert obj.repo_id == repo_uuid
            assert obj.author_email == "test@example.com"
            assert obj.author_name == "Test Author"
            assert obj.commit_hash == "abc123"
            assert obj.line_no == 1
            assert obj.line == "line content"
            assert obj.path == "README.md"


class TestRepoInstanceReuse:
    """Integration tests for repo instance reuse across multiple files."""

    def test_single_repo_instance_for_multiple_files(
        self, repo_path, repo_uuid, git_repo
    ):
        """Test that a single repo instance can be used for multiple files."""
        # Get a few files from the repo
        test_files = [
            os.path.join(repo_path, "README.md"),
            os.path.join(repo_path, "storage.py"),
        ]

        # Process each file with the same repo instance
        all_results = []
        for test_file in test_files:
            if os.path.exists(test_file):
                results = GitBlame.process_file(
                    repo_path, test_file, repo_uuid, repo=git_repo
                )
                all_results.extend(results)

        # Should have processed files successfully
        assert len(all_results) > 0

        # All results should have the same repo_uuid
        for result in all_results:
            assert result.repo_id == repo_uuid



from __future__ import annotations

from datetime import date, datetime, timedelta, timezone

from metrics.compute_work_items import compute_work_item_metrics_daily
from models.work_items import WorkItem


def test_unassigned_completed_items_are_tracked() -> None:
    day = date(2025, 12, 18)
    start = datetime(2025, 12, 18, tzinfo=timezone.utc)
    computed_at = start + timedelta(hours=1)

    assigned = WorkItem(
        work_item_id="jira:ABC-1",
        provider="jira",
        project_key="ABC",
        project_id="1",
        title="Assigned",
        type="task",
        status="done",
        status_raw="Done",
        assignees=["alice@example.com"],
        reporter=None,
        created_at=start - timedelta(days=2),
        updated_at=start,
        started_at=start - timedelta(hours=5),
        completed_at=start + timedelta(hours=1),
        closed_at=start + timedelta(hours=1),
        labels=[],
    )

    unassigned = WorkItem(
        work_item_id="jira:ABC-2",
        provider="jira",
        project_key="ABC",
        project_id="1",
        title="Unassigned",
        type="task",
        status="done",
        status_raw="Done",
        assignees=[],
        reporter=None,
        created_at=start - timedelta(days=2),
        updated_at=start,
        started_at=start - timedelta(hours=2),
        completed_at=start + timedelta(hours=2),
        closed_at=start + timedelta(hours=2),
        labels=[],
    )

    group_rows, user_rows, _cycle_rows = compute_work_item_metrics_daily(
        day=day,
        work_items=[assigned, unassigned],
        transitions=[],
        computed_at=computed_at,
        team_resolver=None,
    )

    assert len(group_rows) == 1
    group = group_rows[0]
    assert group.team_id == "unassigned"
    assert group.items_completed == 2
    assert group.items_completed_unassigned == 1

    # We emit a pseudo-user row so unassigned closures can be charted.
    unassigned_user = next(r for r in user_rows if r.user_identity == "unassigned")
    assert unassigned_user.items_completed == 1



"""
Tests for GitLab connector new features.
"""

from unittest.mock import Mock, patch

import pytest

from connectors import GitLabConnector


class TestGitLabConnectorProjects:
    """Test GitLab connector project listing features."""

    @pytest.fixture
    def mock_gitlab_client(self):
        """Create a mock GitLab client."""
        with patch("connectors.gitlab.gitlab.Gitlab") as mock_gitlab:
            yield mock_gitlab

    @pytest.fixture
    def mock_rest_client(self):
        """Create a mock REST client."""
        with patch("connectors.gitlab.GitLabRESTClient") as mock_rest:
            yield mock_rest

    def test_list_projects_by_group_name(self, mock_gitlab_client, mock_rest_client):
        """Test listing projects for a group by name."""
        # Setup mock
        mock_project = Mock()
        mock_project.id = 1
        mock_project.name = "test-project"
        mock_project.path_with_namespace = "mygroup/test-project"
        mock_project.default_branch = "main"
        mock_project.description = "Test project"
        mock_project.web_url = "https://gitlab.com/mygroup/test-project"
        mock_project.created_at = "2023-01-01T00:00:00Z"
        mock_project.last_activity_at = "2023-01-02T00:00:00Z"
        mock_project.star_count = 10
        mock_project.forks_count = 5

        mock_group = Mock()
        mock_group.projects = Mock()
        mock_group.projects.list.return_value = [mock_project]

        mock_gitlab_instance = mock_gitlab_client.return_value
        mock_gitlab_instance.groups = Mock()
        mock_gitlab_instance.groups.get.return_value = mock_group

        # Test
        connector = GitLabConnector(
            url="https://gitlab.com", private_token="test_token"
        )
        projects = connector.list_projects(group_name="mygroup", max_projects=10)

        # Assert
        assert len(projects) == 1
        assert projects[0].name == "test-project"
        assert projects[0].full_name == "mygroup/test-project"
        mock_gitlab_instance.groups.get.assert_called_once_with("mygroup")

    def test_list_projects_with_search(self, mock_gitlab_client, mock_rest_client):
        """Test listing projects with search query."""
        # Setup mock
        mock_project = Mock()
        mock_project.id = 1
        mock_project.name = "docker-service"
        mock_project.path_with_namespace = "user/docker-service"
        mock_project.default_branch = "main"
        mock_project.description = "Docker service"
        mock_project.web_url = "https://gitlab.com/user/docker-service"
        mock_project.created_at = "2023-01-01T00:00:00Z"
        mock_project.last_activity_at = "2023-01-02T00:00:00Z"
        mock_project.star_count = 15
        mock_project.forks_count = 8

        mock_gitlab_instance = mock_gitlab_client.return_value
        mock_gitlab_instance.projects = Mock()
        mock_gitlab_instance.projects.list.return_value = [mock_project]

        # Test
        connector = GitLabConnector(
            url="https://gitlab.com", private_token="test_token"
        )
        projects = connector.list_projects(search="docker", max_projects=10)

        # Assert
        assert len(projects) == 1
        assert projects[0].name == "docker-service"
        mock_gitlab_instance.projects.list.assert_called_once()
        call_kwargs = mock_gitlab_instance.projects.list.call_args[1]
        assert call_kwargs["search"] == "docker"

    def test_list_projects_search_within_group(
        self, mock_gitlab_client, mock_rest_client
    ):
        """Test searching projects within a group."""
        # Setup mock
        mock_project = Mock()
        mock_project.id = 1
        mock_project.name = "api-backend"
        mock_project.path_with_namespace = "mygroup/api-backend"
        mock_project.default_branch = "main"
        mock_project.description = "API backend"
        mock_project.web_url = "https://gitlab.com/mygroup/api-backend"
        mock_project.created_at = "2023-01-01T00:00:00Z"
        mock_project.last_activity_at = "2023-01-02T00:00:00Z"
        mock_project.star_count = 20
        mock_project.forks_count = 10

        mock_group = Mock()
        mock_group.projects = Mock()
        mock_group.projects.list.return_value = [mock_project]

        mock_gitlab_instance = mock_gitlab_client.return_value
        mock_gitlab_instance.groups = Mock()
        mock_gitlab_instance.groups.get.return_value = mock_group

        # Test
        connector = GitLabConnector(
            url="https://gitlab.com", private_token="test_token"
        )
        projects = connector.list_projects(
            group_name="mygroup", search="api", max_projects=10
        )

        # Assert
        assert len(projects) == 1
        assert projects[0].name == "api-backend"
        mock_gitlab_instance.groups.get.assert_called_once_with("mygroup")
        call_kwargs = mock_group.projects.list.call_args[1]
        assert call_kwargs["search"] == "api"

    def test_list_all_projects_no_limit(self, mock_gitlab_client, mock_rest_client):
        """Test listing all projects without max_projects limit."""
        # Setup mock
        mock_projects = []
        for i in range(150):  # More than typical page size
            mock_project = Mock()
            mock_project.id = i
            mock_project.name = f"project-{i}"
            mock_project.path_with_namespace = f"user/project-{i}"
            mock_project.default_branch = "main"
            mock_project.description = f"Project {i}"
            mock_project.web_url = f"https://gitlab.com/user/project-{i}"
            mock_project.created_at = "2023-01-01T00:00:00Z"
            mock_project.last_activity_at = "2023-01-02T00:00:00Z"
            mock_project.star_count = i
            mock_project.forks_count = i // 2
            mock_projects.append(mock_project)

        mock_gitlab_instance = mock_gitlab_client.return_value
        mock_gitlab_instance.projects = Mock()
        mock_gitlab_instance.projects.list.return_value = mock_projects

        # Test - no max_projects specified
        connector = GitLabConnector(
            url="https://gitlab.com", private_token="test_token"
        )
        projects = connector.list_projects()

        # Assert - all projects should be returned
        assert len(projects) == 150
        # Verify get_all was set to True
        call_kwargs = mock_gitlab_instance.projects.list.call_args[1]
        assert call_kwargs["get_all"] is True

    def test_list_projects_backward_compatibility_group_id(
        self, mock_gitlab_client, mock_rest_client
    ):
        """Test backward compatibility with group_id parameter."""
        # Setup mock
        mock_project = Mock()
        mock_project.id = 1
        mock_project.name = "legacy-project"
        mock_project.path_with_namespace = "mygroup/legacy-project"
        mock_project.default_branch = "main"
        mock_project.description = "Legacy project"
        mock_project.web_url = "https://gitlab.com/mygroup/legacy-project"
        mock_project.created_at = "2023-01-01T00:00:00Z"
        mock_project.last_activity_at = "2023-01-02T00:00:00Z"
        mock_project.star_count = 5
        mock_project.forks_count = 2

        mock_group = Mock()
        mock_group.projects = Mock()
        mock_group.projects.list.return_value = [mock_project]

        mock_gitlab_instance = mock_gitlab_client.return_value
        mock_gitlab_instance.groups = Mock()
        mock_gitlab_instance.groups.get.return_value = mock_group

        # Test - using old group_id parameter
        connector = GitLabConnector(
            url="https://gitlab.com", private_token="test_token"
        )
        projects = connector.list_projects(group_id=123)

        # Assert
        assert len(projects) == 1
        assert projects[0].name == "legacy-project"
        mock_gitlab_instance.groups.get.assert_called_once_with(123)

    def test_get_contributors_with_project_name(
        self, mock_gitlab_client, mock_rest_client
    ):
        """Test getting contributors using project name."""
        # Setup mock
        mock_project = Mock()
        mock_project.repository_contributors.return_value = [
            {"name": "User 1", "email": "user1@example.com"},
            {"name": "User 2", "email": "user2@example.com"},
        ]

        mock_gitlab_instance = mock_gitlab_client.return_value
        mock_gitlab_instance.projects = Mock()
        mock_gitlab_instance.projects.get.return_value = mock_project

        # Test
        connector = GitLabConnector(
            url="https://gitlab.com", private_token="test_token"
        )
        contributors = connector.get_contributors_by_project(
            project_name="mygroup/myproject", max_contributors=10
        )

        # Assert
        assert len(contributors) == 2
        assert contributors[0].username == "User 1"
        mock_gitlab_instance.projects.get.assert_called_once_with("mygroup/myproject")

    def test_get_merge_requests_with_project_name(
        self, mock_gitlab_client, mock_rest_client
    ):
        """Test getting merge requests using project name."""
        # Setup mock
        mock_project = Mock()
        mock_project.id = 123

        mock_gitlab_instance = mock_gitlab_client.return_value
        mock_gitlab_instance.projects = Mock()
        mock_gitlab_instance.projects.get.return_value = mock_project

        mock_rest_instance = mock_rest_client.return_value
        # First call returns MRs, second call returns empty list to stop pagination
        mock_rest_instance.get_merge_requests.side_effect = [
            [
                {
                    "id": 1,
                    "iid": 10,
                    "title": "Test MR",
                    "state": "opened",
                    "author": {
                        "id": 1,
                        "username": "user1",
                        "name": "User 1",
                        "web_url": "https://gitlab.com/user1",
                    },
                    "created_at": "2023-01-01T00:00:00Z",
                    "target_branch": "main",
                    "source_branch": "feature",
                }
            ],
            [],  # Empty list to stop pagination
        ]

        # Test
        connector = GitLabConnector(
            url="https://gitlab.com", private_token="test_token"
        )
        mrs = connector.get_merge_requests(
            project_name="mygroup/myproject", state="opened", max_mrs=10
        )

        # Assert
        assert len(mrs) == 1
        assert mrs[0].title == "Test MR"
        mock_gitlab_instance.projects.get.assert_called_once_with("mygroup/myproject")
        assert mock_rest_instance.get_merge_requests.call_count == 2

    def test_methods_require_project_identifier(
        self, mock_gitlab_client, mock_rest_client
    ):
        """Test that methods raise ValueError when no identifier provided."""
        connector = GitLabConnector(
            url="https://gitlab.com", private_token="test_token"
        )

        error_msg = "Either project_id or project_name must be provided"
        with pytest.raises(ValueError, match=error_msg):
            connector.get_contributors_by_project()

        with pytest.raises(ValueError, match=error_msg):
            connector.get_commit_stats_by_project(sha="abc123")

        with pytest.raises(ValueError, match=error_msg):
            connector.get_repo_stats_by_project()


class TestGitLabConnectorAdapterMethods:
    """Tests for GitLab connector adapter methods (base class interface)."""

    @pytest.fixture
    def mock_gitlab_client(self):
        """Create a mock GitLab client."""
        with patch("connectors.gitlab.gitlab.Gitlab") as mock_gitlab:
            yield mock_gitlab

    @pytest.fixture
    def mock_rest_client(self):
        """Create a mock REST client."""
        with patch("connectors.gitlab.GitLabRESTClient") as mock_rest:
            yield mock_rest

    def test_get_contributors_adapter(self, mock_gitlab_client, mock_rest_client):
        """Test get_contributors adapter method formats project_name correctly."""
        mock_project = Mock()
        mock_project.id = 123
        mock_project.repository_contributors.return_value = []

        mock_gitlab_instance = mock_gitlab_client.return_value
        mock_gitlab_instance.projects = Mock()
        mock_gitlab_instance.projects.get.return_value = mock_project

        connector = GitLabConnector(
            url="https://gitlab.com", private_token="test_token"
        )
        result = connector.get_contributors("mygroup", "myproject")

        # Verify project_name was formatted correctly
        mock_gitlab_instance.projects.get.assert_called_once_with("mygroup/myproject")
        assert result == []

    def test_get_commit_stats_adapter(self, mock_gitlab_client, mock_rest_client):
        """Test get_commit_stats adapter method formats project_name correctly."""
        mock_commit = Mock()
        mock_commit.id = "abc123"
        mock_commit.stats = {"additions": 10, "deletions": 5, "total": 15}

        mock_project = Mock()
        mock_project.id = 123
        mock_project.commits = Mock()
        mock_project.commits.get.return_value = mock_commit

        mock_gitlab_instance = mock_gitlab_client.return_value
        mock_gitlab_instance.projects = Mock()
        mock_gitlab_instance.projects.get.return_value = mock_project

        connector = GitLabConnector(
            url="https://gitlab.com", private_token="test_token"
        )
        result = connector.get_commit_stats("owner", "repo", sha="abc123")

        # Verify project_name was formatted correctly
        mock_gitlab_instance.projects.get.assert_called_once_with("owner/repo")
        assert result.additions == 10
        assert result.deletions == 5

    def test_get_repo_stats_adapter(self, mock_gitlab_client, mock_rest_client):
        """Test get_repo_stats adapter method formats project_name correctly."""
        mock_commit = Mock()
        mock_commit.id = "abc123"
        mock_commit.author_name = "Test Author"
        mock_commit.author_email = "test@example.com"
        mock_commit.committed_date = "2023-01-01T00:00:00.000Z"
        mock_commit.stats = {"additions": 10, "deletions": 5, "total": 15}

        mock_project = Mock()
        mock_project.id = 123
        mock_project.default_branch = "main"
        mock_project.created_at = "2023-01-01T00:00:00.000Z"
        mock_project.commits = Mock()
        mock_project.commits.list.return_value = [mock_commit]
        mock_project.commits.get.return_value = mock_commit
        mock_project.repository_contributors.return_value = []

        mock_gitlab_instance = mock_gitlab_client.return_value
        mock_gitlab_instance.projects = Mock()
        mock_gitlab_instance.projects.get.return_value = mock_project

        connector = GitLabConnector(
            url="https://gitlab.com", private_token="test_token"
        )
        result = connector.get_repo_stats("mygroup", "myproject", max_commits=10)

        # Verify project_name was formatted correctly
        mock_gitlab_instance.projects.get.assert_called_with("mygroup/myproject")
        assert result.total_commits == 1

    def test_get_pull_requests_adapter(self, mock_gitlab_client, mock_rest_client):
        """Test get_pull_requests adapter method maps to get_merge_requests."""
        mock_project = Mock()
        mock_project.id = 123

        mock_gitlab_instance = mock_gitlab_client.return_value
        mock_gitlab_instance.projects = Mock()
        mock_gitlab_instance.projects.get.return_value = mock_project

        mock_rest_instance = mock_rest_client.return_value
        mock_rest_instance.get_merge_requests.return_value = []

        connector = GitLabConnector(
            url="https://gitlab.com", private_token="test_token"
        )
        result = connector.get_pull_requests("owner", "repo", state="open", max_prs=5)

        # Verify project_name was formatted correctly
        mock_gitlab_instance.projects.get.assert_called_once_with("owner/repo")
        assert result == []

    def test_get_file_blame_adapter(self, mock_gitlab_client, mock_rest_client):
        """Test get_file_blame adapter method formats project_name correctly."""
        mock_project = Mock()
        mock_project.id = 123

        mock_gitlab_instance = mock_gitlab_client.return_value
        mock_gitlab_instance.projects = Mock()
        mock_gitlab_instance.projects.get.return_value = mock_project

        mock_rest_instance = mock_rest_client.return_value
        mock_rest_instance.get_file_blame.return_value = []

        connector = GitLabConnector(
            url="https://gitlab.com", private_token="test_token"
        )
        result = connector.get_file_blame(
            "mygroup", "myproject", "path/to/file.py", ref="main"
        )

        # Verify project_name was formatted correctly
        mock_gitlab_instance.projects.get.assert_called_once_with("mygroup/myproject")
        assert result.file_path == "path/to/file.py"
